@@ -34,6 +34,7 @@
 #include "distributed/master_metadata_utility.h"
 #include "distributed/metadata_cache.h"
 #include "distributed/multi_executor.h"
+#include "distributed/multi_partitioning_utils.h"
 #include "distributed/multi_physical_planner.h"
 #include "distributed/multi_planner.h"
 #include "distributed/multi_router_executor.h"
@@ -70,21 +71,22 @@
 
 /* controls use of locks to enforce safe commutativity */
 bool AllModificationsCommutative = false;
+
+/* we've deprecated this flag, keeping here for some time not to break existing users */
 bool EnableDeadlockPrevention = true;
 
 /* functions needed during run phase */
-static void ReacquireMetadataLocks(List *taskList);
-static void AssignInsertTaskShardId(Query *jobQuery, List *taskList);
+static void AcquireMetadataLocks(List *taskList);
+static ShardPlacementAccess * CreatePlacementAccess(ShardPlacement *placement,
+													ShardPlacementAccessType accessType);
 static void ExecuteSingleModifyTask(CitusScanState *scanState, Task *task,
-									bool expectResults);
+									bool multipleTasks, bool expectResults);
 static void ExecuteSingleSelectTask(CitusScanState *scanState, Task *task);
-static List * GetModifyConnections(List *taskPlacementList, bool markCritical,
-								   bool startedInTransaction);
+static List * GetModifyConnections(Task *task, bool markCritical);
 static void ExecuteMultipleTasks(CitusScanState *scanState, List *taskList,
 								 bool isModificationQuery, bool expectResults);
 static int64 ExecuteModifyTasks(List *taskList, bool expectResults,
 								ParamListInfo paramListInfo, CitusScanState *scanState);
-static List * TaskShardIntervalList(List *taskList);
 static void AcquireExecutorShardLock(Task *task, CmdType commandType);
 static void AcquireExecutorMultiShardLocks(List *taskList);
 static bool RequiresConsistentSnapshot(Task *task);
@@ -97,25 +99,16 @@ static bool StoreQueryResult(CitusScanState *scanState, MultiConnection *connect
 							 bool failOnError, int64 *rows);
 static bool ConsumeQueryResult(MultiConnection *connection, bool failOnError,
 							   int64 *rows);
+static LOCKMODE LockModeForModifyTask(Task *task);
 
 
 /*
- * ReacquireMetadataLocks re-acquires the metadata locks that are normally
- * acquired during planning.
- *
- * If we are executing a prepared statement, then planning might have
- * happened in a separate transaction and advisory locks are no longer
- * held. If a shard is currently being repaired/copied/moved, then
- * obtaining the locks will fail and this function throws an error to
- * prevent executing a stale plan.
- *
- * If we are executing a non-prepared statement or planning happened in
- * the same transaction, then we already have the locks and obtain them
- * again here. Since we always release these locks at the end of the
- * transaction, this is effectively a no-op.
+ * AcquireMetadataLocks acquires metadata locks on each of the anchor
+ * shards in the task list to prevent a shard being modified while it
+ * is being copied.
  */
 static void
-ReacquireMetadataLocks(List *taskList)
+AcquireMetadataLocks(List *taskList)
 {
 	ListCell *taskCell = NULL;
 
@@ -130,28 +123,7 @@ ReacquireMetadataLocks(List *taskList)
 	{
 		Task *task = (Task *) lfirst(taskCell);
 
-		/*
-		 * Only obtain metadata locks for modifications to allow reads to
-		 * proceed during shard copy.
-		 */
-		if (task->taskType == MODIFY_TASK &&
-			!TryLockShardDistributionMetadata(task->anchorShardId, ShareLock))
-		{
-			/*
-			 * We could error out immediately to give quick feedback to the
-			 * client, but this might complicate flow control and our default
-			 * behaviour during shard copy is to block.
-			 *
-			 * Block until the lock becomes available such that the next command
-			 * will likely succeed and use the serialization failure error code
-			 * to signal to the client that it should retry the current command.
-			 */
-			LockShardDistributionMetadata(task->anchorShardId, ShareLock);
-
-			ereport(ERROR, (errcode(ERRCODE_T_R_SERIALIZATION_FAILURE),
-							errmsg("prepared modifications cannot be executed on "
-								   "a shard while it is being copied")));
-		}
+		LockShardDistributionMetadata(task->anchorShardId, ShareLock);
 	}
 }
 
@@ -326,6 +298,11 @@ AcquireExecutorMultiShardLocks(List *taskList)
 			lockMode = ExclusiveLock;
 		}
 
+		/*
+		 * If we are dealing with a partition we are also taking locks on parent table
+		 * to prevent deadlocks on concurrent operations on a partition and its parent.
+		 */
+		LockParentShardResourceIfPartition(task->anchorShardId, lockMode);
 		LockShardResource(task->anchorShardId, lockMode);
 
 		/*
@@ -426,84 +403,56 @@ CitusModifyBeginScan(CustomScanState *node, EState *estate, int eflags)
 	if (workerJob->requiresMasterEvaluation)
 	{
 		PlanState *planState = &(scanState->customScanState.ss.ps);
+		EState *executorState = planState->state;
 
 		ExecuteMasterEvaluableFunctions(jobQuery, planState);
 
+		/*
+		 * We've processed parameters in ExecuteMasterEvaluableFunctions and
+		 * don't need to send their values to workers, since they will be
+		 * represented as constants in the deparsed query. To avoid sending
+		 * parameter values, we set the parameter list to NULL.
+		 */
+		executorState->es_param_list_info = NULL;
+
 		if (deferredPruning)
 		{
-			AssignInsertTaskShardId(jobQuery, taskList);
+			DeferredErrorMessage *planningError = NULL;
+
+			/* need to perform shard pruning, rebuild the task list from scratch */
+			taskList = RouterInsertTaskList(jobQuery, &planningError);
+
+			if (planningError != NULL)
+			{
+				RaiseDeferredError(planningError, ERROR);
+			}
+
+			workerJob->taskList = taskList;
 		}
 
 		RebuildQueryStrings(jobQuery, taskList);
 	}
 
-	/*
-	 * If we are executing a prepared statement, then we may not yet have obtained
-	 * the metadata locks in this transaction. To prevent a concurrent shard copy,
-	 * we re-obtain them here or error out if a shard copy has already started.
-	 *
-	 * If a shard copy finishes in between fetching a plan from cache and
-	 * re-acquiring the locks, then we might still run a stale plan, which could
-	 * cause shard placements to diverge. To minimize this window, we take the
-	 * locks as early as possible.
-	 */
-	ReacquireMetadataLocks(taskList);
+	/* prevent concurrent placement changes */
+	AcquireMetadataLocks(taskList);
 
 	/*
-	 * If we deferred shard pruning to the executor, then we still need to assign
-	 * shard placements. We do this after acquiring the metadata locks to ensure
-	 * we can't get stale metadata. At some point, we may want to load all
-	 * placement metadata here.
+	 * We are taking locks on partitions of partitioned tables. These locks are
+	 * necessary for locking tables that appear in the SELECT part of the query.
 	 */
-	if (deferredPruning)
-	{
-		/* modify tasks are always assigned using first-replica policy */
-		workerJob->taskList = FirstReplicaAssignTaskList(taskList);
-	}
+	LockPartitionsInRelationList(multiPlan->relationIdList, AccessShareLock);
+
+	/* modify tasks are always assigned using first-replica policy */
+	workerJob->taskList = FirstReplicaAssignTaskList(taskList);
 }
 
 
 /*
- * AssignInsertTaskShardId performs shard pruning for an insert and sets
- * anchorShardId accordingly.
- */
-static void
-AssignInsertTaskShardId(Query *jobQuery, List *taskList)
-{
-	ShardInterval *shardInterval = NULL;
-	Task *insertTask = NULL;
-	DeferredErrorMessage *planningError = NULL;
-
-	Assert(jobQuery->commandType == CMD_INSERT);
-
-	/*
-	 * We skipped shard pruning in the planner because the partition
-	 * column contained an expression. Perform shard pruning now.
-	 */
-	shardInterval = FindShardForInsert(jobQuery, &planningError);
-	if (planningError != NULL)
-	{
-		RaiseDeferredError(planningError, ERROR);
-	}
-	else if (shardInterval == NULL)
-	{
-		/* expression could not be evaluated */
-		ereport(ERROR, (errmsg("parameters in the partition column are not "
-							   "allowed")));
-	}
-
-	/* assign a shard ID to the task */
-	insertTask = (Task *) linitial(taskList);
-	insertTask->anchorShardId = shardInterval->shardId;
-}
-
-
-/*
- * RouterSingleModifyExecScan executes a single modification query on a
- * distributed plan and returns results if there is any.
+ * RouterSequentialModifyExecScan executes 0 or more modifications on a
+ * distributed table sequentially and returns results if there are any.
  */
 TupleTableSlot *
-RouterSingleModifyExecScan(CustomScanState *node)
+RouterSequentialModifyExecScan(CustomScanState *node)
 {
 	CitusScanState *scanState = (CitusScanState *) node;
 	TupleTableSlot *resultSlot = NULL;
@@ -514,9 +463,26 @@ RouterSingleModifyExecScan(CustomScanState *node)
 		bool hasReturning = multiPlan->hasReturning;
 		Job *workerJob = multiPlan->workerJob;
 		List *taskList = workerJob->taskList;
-		Task *task = (Task *) linitial(taskList);
+		ListCell *taskCell = NULL;
+		bool multipleTasks = list_length(taskList) > 1;
 
-		ExecuteSingleModifyTask(scanState, task, hasReturning);
+		/*
+		 * We could naturally handle function-based transactions (i.e. those using
+		 * PL/pgSQL or similar) by checking the type of queryDesc->dest, but some
+		 * customers already use functions that touch multiple shards from within
+		 * a function, so we'll ignore functions for now.
+		 */
+		if (IsTransactionBlock() || multipleTasks)
+		{
+			BeginOrContinueCoordinatedTransaction();
+		}
+
+		foreach(taskCell, taskList)
+		{
+			Task *task = (Task *) lfirst(taskCell);
+
+			ExecuteSingleModifyTask(scanState, task, multipleTasks, hasReturning);
+		}
 
 		scanState->finishedRemoteScan = true;
 	}
@@ -573,9 +539,16 @@ RouterSelectExecScan(CustomScanState *node)
 		MultiPlan *multiPlan = scanState->multiPlan;
 		Job *workerJob = multiPlan->workerJob;
 		List *taskList = workerJob->taskList;
-		Task *task = (Task *) linitial(taskList);
 
-		ExecuteSingleSelectTask(scanState, task);
+		/* we are taking locks on partitions of partitioned tables */
+		LockPartitionsInRelationList(multiPlan->relationIdList, AccessShareLock);
+
+		if (list_length(taskList) > 0)
+		{
+			Task *task = (Task *) linitial(taskList);
+
+			ExecuteSingleSelectTask(scanState, task);
+		}
 
 		scanState->finishedRemoteScan = true;
 	}
@@ -601,13 +574,7 @@ ExecuteSingleSelectTask(CitusScanState *scanState, Task *task)
 	List *taskPlacementList = task->taskPlacementList;
 	ListCell *taskPlacementCell = NULL;
 	char *queryString = task->queryString;
-
-	if (XactModificationLevel == XACT_MODIFICATION_MULTI_SHARD)
-	{
-		ereport(ERROR, (errcode(ERRCODE_ACTIVE_SQL_TRANSACTION),
-						errmsg("single-shard query may not appear in transaction blocks "
-							   "which contain multi-shard data modifications")));
-	}
+	List *relationShardList = task->relationShardList;
 
 	/*
 	 * Try to run the query to completion on one placement. If the query fails
@@ -620,8 +587,31 @@ ExecuteSingleSelectTask(CitusScanState *scanState, Task *task)
 		bool dontFailOnError = false;
 		int64 currentAffectedTupleCount = 0;
 		int connectionFlags = SESSION_LIFESPAN;
-		MultiConnection *connection =
-			GetPlacementConnection(connectionFlags, taskPlacement, NULL);
+		List *placementAccessList = NIL;
+		MultiConnection *connection = NULL;
+
+		if (list_length(relationShardList) > 0)
+		{
+			placementAccessList = BuildPlacementSelectList(taskPlacement->groupId,
+														   relationShardList);
+		}
+		else
+		{
+			/*
+			 * When the SELECT prunes down to 0 shards, just use the dummy placement.
+			 *
+			 * FIXME: it would be preferable to evaluate the SELECT locally since no
+			 * data from the workers is required.
+			 */
+
+			ShardPlacementAccess *placementAccess =
+				CreatePlacementAccess(taskPlacement, PLACEMENT_ACCESS_SELECT);
+
+			placementAccessList = list_make1(placementAccess);
+		}
+
+		connection = GetPlacementListConnection(connectionFlags, placementAccessList,
+												NULL);
 
 		queryOK = SendQueryInSingleRowMode(connection, queryString, paramListInfo);
 		if (!queryOK)
@@ -641,6 +631,55 @@ ExecuteSingleSelectTask(CitusScanState *scanState, Task *task)
 }
 
 
+/*
+ * BuildPlacementSelectList builds a list of SELECT placement accesses
+ * which can be used to call StartPlacementListConnection or
+ * GetPlacementListConnection.
+ */
+List *
+BuildPlacementSelectList(uint32 groupId, List *relationShardList)
+{
+	ListCell *relationShardCell = NULL;
+	List *placementAccessList = NIL;
+
+	foreach(relationShardCell, relationShardList)
+	{
+		RelationShard *relationShard = (RelationShard *) lfirst(relationShardCell);
+		ShardPlacement *placement = NULL;
+		ShardPlacementAccess *placementAccess = NULL;
+
+		placement = FindShardPlacementOnGroup(groupId, relationShard->shardId);
+		if (placement == NULL)
+		{
+			ereport(ERROR, (errmsg("no active placement of shard %ld found on group %d",
+								   relationShard->shardId, groupId)));
+		}
+
+		placementAccess = CreatePlacementAccess(placement, PLACEMENT_ACCESS_SELECT);
+		placementAccessList = lappend(placementAccessList, placementAccess);
+	}
+
+	return placementAccessList;
+}
+
+
+/*
+ * CreatePlacementAccess returns a new ShardPlacementAccess for the given placement
+ * and access type.
+ */
+static ShardPlacementAccess *
+CreatePlacementAccess(ShardPlacement *placement, ShardPlacementAccessType accessType)
+{
+	ShardPlacementAccess *placementAccess = NULL;
+
+	placementAccess = (ShardPlacementAccess *) palloc0(sizeof(ShardPlacementAccess));
+	placementAccess->placement = placement;
+	placementAccess->accessType = accessType;
+
+	return placementAccess;
+}
+
+
 /*
  * ExecuteSingleModifyTask executes the task on the remote node, retrieves the
  * results and stores them, if RETURNING is used, in a tuple store.
@@ -651,7 +690,8 @@ ExecuteSingleSelectTask(CitusScanState *scanState, Task *task)
  * framework), or errors out (failed on all placements).
  */
 static void
-ExecuteSingleModifyTask(CitusScanState *scanState, Task *task, bool expectResults)
+ExecuteSingleModifyTask(CitusScanState *scanState, Task *task, bool multipleTasks,
+						bool expectResults)
 {
 	CmdType operation = scanState->multiPlan->operation;
 	EState *executorState = scanState->customScanState.ss.ps.state;
@@ -666,16 +706,9 @@ ExecuteSingleModifyTask(CitusScanState *scanState, Task *task, bool expectResult
 
 	char *queryString = task->queryString;
 	bool taskRequiresTwoPhaseCommit = (task->replicationModel == REPLICATION_MODEL_2PC);
-	bool startedInTransaction =
-		InCoordinatedTransaction() && XactModificationLevel == XACT_MODIFICATION_DATA;
 
-	if (XactModificationLevel == XACT_MODIFICATION_MULTI_SHARD)
-	{
-		ereport(ERROR, (errcode(ERRCODE_ACTIVE_SQL_TRANSACTION),
-						errmsg("single-shard DML commands must not appear in "
-							   "transaction blocks which contain multi-shard data "
-							   "modifications")));
-	}
+	ShardInterval *shardInterval = LoadShardInterval(task->anchorShardId);
+	Oid relationId = shardInterval->relationId;
 
 	/*
 	 * Modifications for reference tables are always done using 2PC. First
@@ -689,25 +722,23 @@ ExecuteSingleModifyTask(CitusScanState *scanState, Task *task, bool expectResult
 		CoordinatedTransactionUse2PC();
 	}
 
-	/*
-	 * We could naturally handle function-based transactions (i.e. those using
-	 * PL/pgSQL or similar) by checking the type of queryDesc->dest, but some
-	 * customers already use functions that touch multiple shards from within
-	 * a function, so we'll ignore functions for now.
-	 */
-	if (IsTransactionBlock())
-	{
-		BeginOrContinueCoordinatedTransaction();
-	}
-
 	/*
 	 * Get connections required to execute task. This will, if necessary,
 	 * establish the connection, mark as critical (when modifying reference
 	 * table) and start a transaction (when in a transaction).
 	 */
-	connectionList = GetModifyConnections(taskPlacementList,
-										  taskRequiresTwoPhaseCommit,
-										  startedInTransaction);
+	connectionList = GetModifyConnections(task, taskRequiresTwoPhaseCommit);
+
+	/*
+	 * If we are dealing with a partitioned table, we also need to lock its
+	 * partitions.
+	 */
+	if (PartitionedTable(relationId))
+	{
+		LOCKMODE lockMode = LockModeForModifyTask(task);
+
+		LockPartitionRelations(relationId, lockMode);
+	}
 
 	/* prevent replicas of the same shard from diverging */
 	AcquireExecutorShardLock(task, operation);
@@ -742,6 +773,15 @@ ExecuteSingleModifyTask(CitusScanState *scanState, Task *task, bool expectResult
 		/* if we're running a 2PC, the query should fail on error */
 		failOnError = taskRequiresTwoPhaseCommit;
 
+		if (multipleTasks && expectResults)
+		{
+			/*
+			 * If we have multiple tasks and one fails, we cannot clear
+			 * the tuple store and start over. Error out instead.
+			 */
+			failOnError = true;
+		}
+
 		/*
 		 * If caller is interested, store query results the first time
 		 * through. The output of the query's execution on other shards is
@@ -775,12 +815,6 @@ ExecuteSingleModifyTask(CitusScanState *scanState, Task *task, bool expectResult
 								   taskPlacement->nodeName, taskPlacement->nodePort)));
 			}
 
-#if (PG_VERSION_NUM < 90600)
-
-			/* before 9.6, PostgreSQL used a uint32 for this field, so check */
-			Assert(currentAffectedTupleCount <= 0xFFFFFFFF);
-#endif
-
 			resultsOK = true;
 			gotResults = true;
 		}
@@ -795,7 +829,7 @@ ExecuteSingleModifyTask(CitusScanState *scanState, Task *task, bool expectResult
 	/* if some placements failed, ensure future statements don't access them */
 	MarkFailedShardPlacements();
 
-	executorState->es_processed = affectedTupleCount;
+	executorState->es_processed += affectedTupleCount;
 
 	if (IsTransactionBlock())
 	{
@@ -809,15 +843,15 @@ ExecuteSingleModifyTask(CitusScanState *scanState, Task *task, bool expectResult
  * modify commands on the placements in tasPlacementList.  If necessary remote
  * transactions are started.
  *
- * If markCritical is true remote transactions are marked as critical. If
- * noNewTransactions is true, this function errors out if there's no
- * transaction in progress.
+ * If markCritical is true remote transactions are marked as critical.
  */
 static List *
-GetModifyConnections(List *taskPlacementList, bool markCritical, bool noNewTransactions)
+GetModifyConnections(Task *task, bool markCritical)
 {
+	List *taskPlacementList = task->taskPlacementList;
 	ListCell *taskPlacementCell = NULL;
 	List *multiConnectionList = NIL;
+	List *relationShardList = task->relationShardList;
 
 	/* first initiate connection establishment for all necessary connections */
 	foreach(taskPlacementCell, taskPlacementList)
@@ -825,32 +859,34 @@ GetModifyConnections(List *taskPlacementList, bool markCritical, bool noNewTrans
 		ShardPlacement *taskPlacement = (ShardPlacement *) lfirst(taskPlacementCell);
 		int connectionFlags = SESSION_LIFESPAN | FOR_DML;
 		MultiConnection *multiConnection = NULL;
+		List *placementAccessList = NIL;
+		ShardPlacementAccess *placementModification = NULL;
+
+		/* create placement accesses for placements that appear in a subselect */
+		placementAccessList = BuildPlacementSelectList(taskPlacement->groupId,
+													   relationShardList);
+
+		/* create placement access for the placement that we're modifying */
+		placementModification = CreatePlacementAccess(taskPlacement,
+													  PLACEMENT_ACCESS_DML);
+		placementAccessList = lappend(placementAccessList, placementModification);
+
+		/* get an appropriate connection for the DML statement */
+		multiConnection = GetPlacementListConnection(connectionFlags, placementAccessList,
+													 NULL);
 
 		/*
-		 * FIXME: It's not actually correct to use only one shard placement
-		 * here for router queries involving multiple relations. We should
-		 * check that this connection is the only modifying one associated
-		 * with all the involved shards.
+		 * If we're expanding the set nodes that participate in the distributed
+		 * transaction, conform to MultiShardCommitProtocol.
 		 */
-		multiConnection = StartPlacementConnection(connectionFlags, taskPlacement, NULL);
-
-		/*
-		 * If already in a transaction, disallow expanding set of remote
-		 * transactions. That prevents some forms of distributed deadlocks.
-		 */
-		if (noNewTransactions)
+		if (MultiShardCommitProtocol == COMMIT_PROTOCOL_2PC &&
+			InCoordinatedTransaction() &&
+			XactModificationLevel == XACT_MODIFICATION_DATA)
 		{
 			RemoteTransaction *transaction = &multiConnection->remoteTransaction;
-
-			if (EnableDeadlockPrevention &&
-				transaction->transactionState == REMOTE_TRANS_INVALID)
+			if (transaction->transactionState == REMOTE_TRANS_INVALID)
 			{
-				ereport(ERROR, (errcode(ERRCODE_CONNECTION_DOES_NOT_EXIST),
-								errmsg("no transaction participant matches %s:%d",
-									   taskPlacement->nodeName, taskPlacement->nodePort),
-								errdetail("Transactions which modify distributed tables "
-										  "may only target nodes affected by the "
-										  "modification command which began the transaction.")));
+				CoordinatedTransactionUse2PC();
 			}
 		}
 
@@ -948,8 +984,8 @@ ExecuteModifyTasks(List *taskList, bool expectResults, ParamListInfo paramListIn
 	int64 totalAffectedTupleCount = 0;
 	ListCell *taskCell = NULL;
 	Task *firstTask = NULL;
+	ShardInterval *firstShardInterval = NULL;
 	int connectionFlags = 0;
-	List *shardIntervalList = NIL;
 	List *affectedTupleCountList = NIL;
 	HTAB *shardConnectionHash = NULL;
 	bool tasksPending = true;
@@ -960,23 +996,25 @@ ExecuteModifyTasks(List *taskList, bool expectResults, ParamListInfo paramListIn
 		return 0;
 	}
 
-	if (XactModificationLevel == XACT_MODIFICATION_DATA)
+	/*
+	 * In multi shard modification, we expect that all tasks operates on the
+	 * same relation, so it is enough to acquire a lock on the first task's
+	 * anchor relation's partitions.
+	 */
+	firstTask = (Task *) linitial(taskList);
+	firstShardInterval = LoadShardInterval(firstTask->anchorShardId);
+	if (PartitionedTable(firstShardInterval->relationId))
 	{
-		ereport(ERROR, (errcode(ERRCODE_ACTIVE_SQL_TRANSACTION),
-						errmsg("multi-shard data modifications must not appear in "
-							   "transaction blocks which contain single-shard DML "
-							   "commands")));
-	}
+		LOCKMODE lockMode = LockModeForModifyTask(firstTask);
 
-	shardIntervalList = TaskShardIntervalList(taskList);
+		LockPartitionRelations(firstShardInterval->relationId, lockMode);
+	}
 
 	/* ensure that there are no concurrent modifications on the same shards */
 	AcquireExecutorMultiShardLocks(taskList);
 
 	BeginOrContinueCoordinatedTransaction();
 
-	firstTask = (Task *) linitial(taskList);
-
 	if (MultiShardCommitProtocol == COMMIT_PROTOCOL_2PC ||
 		firstTask->replicationModel == REPLICATION_MODEL_2PC)
 	{
@@ -993,10 +1031,9 @@ ExecuteModifyTasks(List *taskList, bool expectResults, ParamListInfo paramListIn
 	}
 
 	/* open connection to all relevant placements, if not already open */
-	shardConnectionHash = OpenTransactionsToAllShardPlacements(shardIntervalList,
-															   connectionFlags);
+	shardConnectionHash = OpenTransactionsForAllTasks(taskList, connectionFlags);
 
-	XactModificationLevel = XACT_MODIFICATION_MULTI_SHARD;
+	XactModificationLevel = XACT_MODIFICATION_DATA;
 
 	/* iterate over placements in rounds, to ensure in-order execution */
 	while (tasksPending)
@@ -1132,29 +1169,6 @@ ExecuteModifyTasks(List *taskList, bool expectResults, ParamListInfo paramListIn
 }
 
 
-/*
- * TaskShardIntervalList returns a list of shard intervals for a given list of
- * tasks.
- */
-static List *
-TaskShardIntervalList(List *taskList)
-{
-	ListCell *taskCell = NULL;
-	List *shardIntervalList = NIL;
-
-	foreach(taskCell, taskList)
-	{
-		Task *task = (Task *) lfirst(taskCell);
-		int64 shardId = task->anchorShardId;
-		ShardInterval *shardInterval = LoadShardInterval(shardId);
-
-		shardIntervalList = lappend(shardIntervalList, shardInterval);
-	}
-
-	return shardIntervalList;
-}
-
-
 /*
  * SendQueryInSingleRowMode sends the given query on the connection in an
  * asynchronous way. The function also sets the single-row mode on the
@@ -1487,11 +1501,6 @@ ConsumeQueryResult(MultiConnection *connection, bool failOnError, int64 *rows)
 				Assert(currentAffectedTupleCount >= 0);
 			}
 
-#if (PG_VERSION_NUM < 90600)
-
-			/* before 9.6, PostgreSQL used a uint32 for this field, so check */
-			Assert(currentAffectedTupleCount <= 0xFFFFFFFF);
-#endif
 			*rows += currentAffectedTupleCount;
 		}
 		else
@@ -1505,3 +1514,29 @@ ConsumeQueryResult(MultiConnection *connection, bool failOnError, int64 *rows)
 
 	return gotResponse && !commandFailed;
 }
+
+
+/*
+ * LockModeForRouterModifyTask returns appropriate LOCKMODE for given router
+ * modify task.
+ */
+static LOCKMODE
+LockModeForModifyTask(Task *task)
+{
+	LOCKMODE lockMode = NoLock;
+	if (task->taskType == DDL_TASK)
+	{
+		lockMode = AccessExclusiveLock;
+	}
+	else if (task->taskType == MODIFY_TASK)
+	{
+		lockMode = RowExclusiveLock;
+	}
+	else
+	{
+		/* we do not allow any other task type in these code path */
+		Assert(false);
+	}
+
+	return lockMode;
+}
