@@ -43,12 +43,10 @@ MultiExecutorType
 JobExecutorType(MultiPlan *multiPlan)
 {
 	Job *job = multiPlan->workerJob;
-	List *workerTaskList = job->taskList;
-	List *workerNodeList = ActiveWorkerNodeList();
-	int taskCount = list_length(workerTaskList);
-	int workerNodeCount = list_length(workerNodeList);
-	double tasksPerNode = taskCount / ((double) workerNodeCount);
-	int dependedJobCount = list_length(job->dependedJobList);
+	List *workerNodeList = NIL;
+	int workerNodeCount = 0;
+	int taskCount = 0;
+	double tasksPerNode = 0.;
 	MultiExecutorType executorType = TaskExecutorType;
 	bool routerExecutablePlan = multiPlan->routerExecutable;
 
@@ -59,6 +57,11 @@ JobExecutorType(MultiPlan *multiPlan)
 		return MULTI_EXECUTOR_ROUTER;
 	}
 
+	if (multiPlan->insertSelectSubquery != NULL)
+	{
+		return MULTI_EXECUTOR_COORDINATOR_INSERT_SELECT;
+	}
+
 	/* if it is not a router executable plan, inform user according to the log level */
 	if (MultiTaskQueryLogLevel != MULTI_TASK_QUERY_INFO_OFF)
 	{
@@ -68,9 +71,17 @@ JobExecutorType(MultiPlan *multiPlan)
 												 " queries on the workers.")));
 	}
 
+	Assert(multiPlan->operation == CMD_SELECT);
+
+	workerNodeList = ActiveReadableNodeList();
+	workerNodeCount = list_length(workerNodeList);
+	taskCount = list_length(job->taskList);
+	tasksPerNode = taskCount / ((double) workerNodeCount);
+
 	if (executorType == MULTI_EXECUTOR_REAL_TIME)
 	{
 		double reasonableConnectionCount = 0;
+		int dependedJobCount = 0;
 
 		/* if we need to open too many connections per worker, warn the user */
 		if (tasksPerNode >= MaxConnections)
@@ -98,6 +109,7 @@ JobExecutorType(MultiPlan *multiPlan)
 		}
 
 		/* if we have repartition jobs with real time executor, error out */
+		dependedJobCount = list_length(job->dependedJobList);
 		if (dependedJobCount > 0)
 		{
 			ereport(ERROR, (errmsg("cannot use real time executor with repartition jobs"),
@@ -160,7 +172,8 @@ InitTaskExecution(Task *task, TaskExecStatus initialTaskExecStatus)
 	uint32 nodeCount = list_length(task->taskPlacementList);
 	uint32 nodeIndex = 0;
 
-	TaskExecution *taskExecution = palloc0(sizeof(TaskExecution));
+	TaskExecution *taskExecution = CitusMakeNode(TaskExecution);
+
 	taskExecution->jobId = task->jobId;
 	taskExecution->taskId = task->taskId;
 	taskExecution->nodeCount = nodeCount;
@@ -225,7 +238,7 @@ CleanupTaskExecution(TaskExecution *taskExecution)
 	pfree(taskExecution->taskStatusArray);
 	pfree(taskExecution->connectionIdArray);
 	pfree(taskExecution->fileDescriptorArray);
-	memset(taskExecution, 0, sizeof(TaskExecution));
+	pfree(taskExecution);
 }
 
 
