@@ -35,7 +35,7 @@
 #include "distributed/pg_dist_colocation.h"
 #include "distributed/pg_dist_partition.h"
 #include "distributed/pg_dist_shard.h"
-#include "distributed/pg_dist_shard_placement.h"
+#include "distributed/pg_dist_placement.h"
 #include "distributed/relay_utility.h"
 #include "distributed/resource_lock.h"
 #include "distributed/remote_commands.h"
@@ -59,12 +59,12 @@
 static uint64 * AllocateUint64(uint64 value);
 static void RecordDistributedRelationDependencies(Oid distributedRelationId,
 												  Node *distributionKey);
-static ShardPlacement * TupleToShardPlacement(TupleDesc tupleDesc,
-											  HeapTuple heapTuple);
+static GroupShardPlacement * TupleToGroupShardPlacement(TupleDesc tupleDesc,
+														HeapTuple heapTuple);
 static uint64 DistributedTableSize(Oid relationId, char *sizeQuery);
 static uint64 DistributedTableSizeOnWorker(WorkerNode *workerNode, Oid relationId,
 										   char *sizeQuery);
-static List * ShardIntervalsOnWorkerNode(WorkerNode *workerNode, Oid relationId);
+static List * ShardIntervalsOnWorkerGroup(WorkerNode *workerNode, Oid relationId);
 static StringInfo GenerateSizeQueryOnMultiplePlacements(Oid distributedRelationId,
 														List *shardIntervalList,
 														char *sizeQuery);
@@ -86,11 +86,16 @@ citus_total_relation_size(PG_FUNCTION_ARGS)
 {
 	Oid relationId = PG_GETARG_OID(0);
 	uint64 totalRelationSize = 0;
+	char *tableSizeFunction = PG_TOTAL_RELATION_SIZE_FUNCTION;
 
 	CheckCitusVersion(ERROR);
 
-	totalRelationSize = DistributedTableSize(relationId,
-											 PG_TOTAL_RELATION_SIZE_FUNCTION);
+	if (CStoreTable(relationId))
+	{
+		tableSizeFunction = CSTORE_TABLE_SIZE_FUNCTION;
+	}
+
+	totalRelationSize = DistributedTableSize(relationId, tableSizeFunction);
 
 	PG_RETURN_INT64(totalRelationSize);
 }
@@ -105,10 +110,16 @@ citus_table_size(PG_FUNCTION_ARGS)
 {
 	Oid relationId = PG_GETARG_OID(0);
 	uint64 tableSize = 0;
+	char *tableSizeFunction = PG_TABLE_SIZE_FUNCTION;
 
 	CheckCitusVersion(ERROR);
 
-	tableSize = DistributedTableSize(relationId, PG_TABLE_SIZE_FUNCTION);
+	if (CStoreTable(relationId))
+	{
+		tableSizeFunction = CSTORE_TABLE_SIZE_FUNCTION;
+	}
+
+	tableSize = DistributedTableSize(relationId, tableSizeFunction);
 
 	PG_RETURN_INT64(tableSize);
 }
@@ -123,10 +134,16 @@ citus_relation_size(PG_FUNCTION_ARGS)
 {
 	Oid relationId = PG_GETARG_OID(0);
 	uint64 relationSize = 0;
+	char *tableSizeFunction = PG_RELATION_SIZE_FUNCTION;
 
 	CheckCitusVersion(ERROR);
 
-	relationSize = DistributedTableSize(relationId, PG_RELATION_SIZE_FUNCTION);
+	if (CStoreTable(relationId))
+	{
+		tableSizeFunction = CSTORE_TABLE_SIZE_FUNCTION;
+	}
+
+	relationSize = DistributedTableSize(relationId, tableSizeFunction);
 
 	PG_RETURN_INT64(relationSize);
 }
@@ -141,12 +158,11 @@ static uint64
 DistributedTableSize(Oid relationId, char *sizeQuery)
 {
 	Relation relation = NULL;
-	Relation pgDistNode = NULL;
 	List *workerNodeList = NULL;
 	ListCell *workerNodeCell = NULL;
 	uint64 totalRelationSize = 0;
 
-	if (XactModificationLevel == XACT_MODIFICATION_MULTI_SHARD)
+	if (XactModificationLevel == XACT_MODIFICATION_DATA)
 	{
 		ereport(ERROR, (errcode(ERRCODE_ACTIVE_SQL_TRANSACTION),
 						errmsg("citus size functions cannot be called in transaction"
@@ -158,9 +174,7 @@ DistributedTableSize(Oid relationId, char *sizeQuery)
 
 	ErrorIfNotSuitableToGetSize(relationId);
 
-	pgDistNode = heap_open(DistNodeRelationId(), AccessShareLock);
-
-	workerNodeList = ActiveWorkerNodeList();
+	workerNodeList = ActiveReadableNodeList();
 
 	foreach(workerNodeCell, workerNodeList)
 	{
@@ -170,7 +184,6 @@ DistributedTableSize(Oid relationId, char *sizeQuery)
 		totalRelationSize += relationSizeOnNode;
 	}
 
-	heap_close(pgDistNode, NoLock);
 	heap_close(relation, AccessShareLock);
 
 	return totalRelationSize;
@@ -190,23 +203,29 @@ DistributedTableSizeOnWorker(WorkerNode *workerNode, Oid relationId, char *sizeQ
 	char *workerNodeName = workerNode->workerName;
 	uint32 workerNodePort = workerNode->workerPort;
 	char *tableSizeString;
-	List *sizeList = NIL;
 	uint64 tableSize = 0;
+	MultiConnection *connection = NULL;
+	uint32 connectionFlag = FORCE_NEW_CONNECTION;
+	PGresult *result = NULL;
+	int queryResult = 0;
+	List *sizeList = NIL;
 
-	List *shardIntervalsOnNode = ShardIntervalsOnWorkerNode(workerNode, relationId);
+	List *shardIntervalsOnNode = ShardIntervalsOnWorkerGroup(workerNode, relationId);
 
 	tableSizeQuery = GenerateSizeQueryOnMultiplePlacements(relationId,
 														   shardIntervalsOnNode,
 														   sizeQuery);
 
-	sizeList = ExecuteRemoteQuery(workerNodeName, workerNodePort, NULL, tableSizeQuery);
+	connection = GetNodeConnection(connectionFlag, workerNodeName, workerNodePort);
+	queryResult = ExecuteOptionalRemoteCommand(connection, tableSizeQuery->data, &result);
 
-	if (sizeList == NIL)
+	if (queryResult != 0)
 	{
 		ereport(ERROR, (errcode(ERRCODE_CONNECTION_FAILURE),
 						errmsg("cannot get the size because of a connection error")));
 	}
 
+	sizeList = ReadFirstColumnAsText(result);
 	tableSizeStringInfo = (StringInfo) linitial(sizeList);
 	tableSizeString = tableSizeStringInfo->data;
 	tableSize = atol(tableSizeString);
@@ -216,26 +235,22 @@ DistributedTableSizeOnWorker(WorkerNode *workerNode, Oid relationId, char *sizeQ
 
 
 /*
- * ShardIntervalsOnNode takes a WorkerNode then compares it with each placement
- * of table. It returns shard intervals of table on that node as a list of shard
- * intervals. Note that, shard intervals returned as elements of the list are
- * not the copies but the pointers.
- *
- * DO NOT modify the shard intervals returned by this function.
+ * GroupShardPlacementsForTableOnGroup accepts a relationId and a group and returns a list
+ * of GroupShardPlacement's representing all of the placements for the table which reside
+ * on the group.
  */
-static List *
-ShardIntervalsOnWorkerNode(WorkerNode *workerNode, Oid relationId)
+List *
+GroupShardPlacementsForTableOnGroup(Oid relationId, uint32 groupId)
 {
 	DistTableCacheEntry *distTableCacheEntry = DistributedTableCacheEntry(relationId);
-	char *workerNodeName = workerNode->workerName;
-	uint32 workerNodePort = workerNode->workerPort;
-	List *shardIntervalList = NIL;
+	List *resultList = NIL;
+
 	int shardIndex = 0;
 	int shardIntervalArrayLength = distTableCacheEntry->shardIntervalArrayLength;
 
 	for (shardIndex = 0; shardIndex < shardIntervalArrayLength; shardIndex++)
 	{
-		ShardPlacement *placementArray =
+		GroupShardPlacement *placementArray =
 			distTableCacheEntry->arrayOfPlacementArrays[shardIndex];
 		int numberOfPlacements =
 			distTableCacheEntry->arrayOfPlacementArrayLengths[shardIndex];
@@ -243,9 +258,45 @@ ShardIntervalsOnWorkerNode(WorkerNode *workerNode, Oid relationId)
 
 		for (placementIndex = 0; placementIndex < numberOfPlacements; placementIndex++)
 		{
-			ShardPlacement *placement = &placementArray[placementIndex];
-			char *shardNodeName = placement->nodeName;
-			uint32 shardNodePort = placement->nodePort;
+			GroupShardPlacement *placement = &placementArray[placementIndex];
+
+			if (placement->groupId == groupId)
+			{
+				resultList = lappend(resultList, placement);
+			}
+		}
+	}
+
+	return resultList;
+}
+
+
+/*
+ * ShardIntervalsOnWorkerGroup accepts a WorkerNode and returns a list of the shard
+ * intervals of the given table which are placed on the group the node is a part of.
+ *
+ * DO NOT modify the shard intervals returned by this function, they are not copies but
+ * pointers.
+ */
+static List *
+ShardIntervalsOnWorkerGroup(WorkerNode *workerNode, Oid relationId)
+{
+	DistTableCacheEntry *distTableCacheEntry = DistributedTableCacheEntry(relationId);
+	List *shardIntervalList = NIL;
+	int shardIndex = 0;
+	int shardIntervalArrayLength = distTableCacheEntry->shardIntervalArrayLength;
+
+	for (shardIndex = 0; shardIndex < shardIntervalArrayLength; shardIndex++)
+	{
+		GroupShardPlacement *placementArray =
+			distTableCacheEntry->arrayOfPlacementArrays[shardIndex];
+		int numberOfPlacements =
+			distTableCacheEntry->arrayOfPlacementArrayLengths[shardIndex];
+		int placementIndex = 0;
+
+		for (placementIndex = 0; placementIndex < numberOfPlacements; placementIndex++)
+		{
+			GroupShardPlacement *placement = &placementArray[placementIndex];
 			uint64 shardId = placement->shardId;
 			bool metadataLock = false;
 
@@ -260,8 +311,7 @@ ShardIntervalsOnWorkerNode(WorkerNode *workerNode, Oid relationId)
 				continue;
 			}
 
-			if (strcmp(shardNodeName, workerNodeName) == 0 &&
-				shardNodePort == workerNodePort)
+			if (placement->groupId == workerNode->groupId)
 			{
 				ShardInterval *shardInterval =
 					distTableCacheEntry->sortedShardIntervalArray[shardIndex];
@@ -569,12 +619,12 @@ ShardLength(uint64 shardId)
 
 
 /*
- * NodeHasActiveShardPlacements returns whether any active shards are placed on this node
+ * NodeGroupHasShardPlacements returns whether any active shards are placed on the group
  */
 bool
-NodeHasActiveShardPlacements(char *nodeName, int32 nodePort)
+NodeGroupHasShardPlacements(uint32 groupId, bool onlyConsiderActivePlacements)
 {
-	const int scanKeyCount = 3;
+	const int scanKeyCount = (onlyConsiderActivePlacements ? 2 : 1);
 	const bool indexOK = false;
 
 	bool hasFinalizedPlacements = false;
@@ -583,25 +633,26 @@ NodeHasActiveShardPlacements(char *nodeName, int32 nodePort)
 	SysScanDesc scanDescriptor = NULL;
 	ScanKeyData scanKey[scanKeyCount];
 
-	Relation pgShardPlacement = heap_open(DistShardPlacementRelationId(),
-										  AccessShareLock);
+	Relation pgPlacement = heap_open(DistPlacementRelationId(),
+									 AccessShareLock);
 
-	ScanKeyInit(&scanKey[0], Anum_pg_dist_shard_placement_nodename,
-				BTEqualStrategyNumber, F_TEXTEQ, CStringGetTextDatum(nodeName));
-	ScanKeyInit(&scanKey[1], Anum_pg_dist_shard_placement_nodeport,
-				BTEqualStrategyNumber, F_INT8EQ, Int64GetDatum(nodePort));
-	ScanKeyInit(&scanKey[2], Anum_pg_dist_shard_placement_shardstate,
-				BTEqualStrategyNumber, F_INT4EQ, Int32GetDatum(FILE_FINALIZED));
+	ScanKeyInit(&scanKey[0], Anum_pg_dist_placement_groupid,
+				BTEqualStrategyNumber, F_INT4EQ, UInt32GetDatum(groupId));
+	if (onlyConsiderActivePlacements)
+	{
+		ScanKeyInit(&scanKey[1], Anum_pg_dist_placement_shardstate,
+					BTEqualStrategyNumber, F_INT4EQ, Int32GetDatum(FILE_FINALIZED));
+	}
 
-	scanDescriptor = systable_beginscan(pgShardPlacement,
-										DistShardPlacementNodeidIndexId(), indexOK,
+	scanDescriptor = systable_beginscan(pgPlacement,
+										DistPlacementGroupidIndexId(), indexOK,
 										NULL, scanKeyCount, scanKey);
 
 	heapTuple = systable_getnext(scanDescriptor);
 	hasFinalizedPlacements = HeapTupleIsValid(heapTuple);
 
 	systable_endscan(scanDescriptor);
-	heap_close(pgShardPlacement, AccessShareLock);
+	heap_close(pgPlacement, NoLock);
 
 	return hasFinalizedPlacements;
 }
@@ -675,76 +726,75 @@ BuildShardPlacementList(ShardInterval *shardInterval)
 {
 	int64 shardId = shardInterval->shardId;
 	List *shardPlacementList = NIL;
-	Relation pgShardPlacement = NULL;
+	Relation pgPlacement = NULL;
 	SysScanDesc scanDescriptor = NULL;
 	ScanKeyData scanKey[1];
 	int scanKeyCount = 1;
 	bool indexOK = true;
 	HeapTuple heapTuple = NULL;
 
-	pgShardPlacement = heap_open(DistShardPlacementRelationId(), AccessShareLock);
+	pgPlacement = heap_open(DistPlacementRelationId(), AccessShareLock);
 
-	ScanKeyInit(&scanKey[0], Anum_pg_dist_shard_placement_shardid,
+	ScanKeyInit(&scanKey[0], Anum_pg_dist_placement_shardid,
 				BTEqualStrategyNumber, F_INT8EQ, Int64GetDatum(shardId));
 
-	scanDescriptor = systable_beginscan(pgShardPlacement,
-										DistShardPlacementShardidIndexId(), indexOK,
+	scanDescriptor = systable_beginscan(pgPlacement,
+										DistPlacementShardidIndexId(), indexOK,
 										NULL, scanKeyCount, scanKey);
 
 	heapTuple = systable_getnext(scanDescriptor);
 	while (HeapTupleIsValid(heapTuple))
 	{
-		TupleDesc tupleDescriptor = RelationGetDescr(pgShardPlacement);
+		TupleDesc tupleDescriptor = RelationGetDescr(pgPlacement);
+
+		GroupShardPlacement *placement =
+			TupleToGroupShardPlacement(tupleDescriptor, heapTuple);
 
-		ShardPlacement *placement = TupleToShardPlacement(tupleDescriptor, heapTuple);
 		shardPlacementList = lappend(shardPlacementList, placement);
 
 		heapTuple = systable_getnext(scanDescriptor);
 	}
 
 	systable_endscan(scanDescriptor);
-	heap_close(pgShardPlacement, AccessShareLock);
+	heap_close(pgPlacement, NoLock);
 
 	return shardPlacementList;
 }
 
 
 /*
- * TupleToShardPlacement takes in a heap tuple from pg_dist_shard_placement,
+ * TupleToGroupShardPlacement takes in a heap tuple from pg_dist_placement,
  * and converts this tuple to in-memory struct. The function assumes the
  * caller already has locks on the tuple, and doesn't perform any locking.
  */
-static ShardPlacement *
-TupleToShardPlacement(TupleDesc tupleDescriptor, HeapTuple heapTuple)
+static GroupShardPlacement *
+TupleToGroupShardPlacement(TupleDesc tupleDescriptor, HeapTuple heapTuple)
 {
-	ShardPlacement *shardPlacement = NULL;
+	GroupShardPlacement *shardPlacement = NULL;
 	bool isNull = false;
 
-	Datum placementId = heap_getattr(heapTuple, Anum_pg_dist_shard_placement_placementid,
+	Datum placementId = heap_getattr(heapTuple, Anum_pg_dist_placement_placementid,
 									 tupleDescriptor, &isNull);
-	Datum shardId = heap_getattr(heapTuple, Anum_pg_dist_shard_placement_shardid,
+	Datum shardId = heap_getattr(heapTuple, Anum_pg_dist_placement_shardid,
 								 tupleDescriptor, &isNull);
-	Datum shardLength = heap_getattr(heapTuple, Anum_pg_dist_shard_placement_shardlength,
+	Datum shardLength = heap_getattr(heapTuple, Anum_pg_dist_placement_shardlength,
 									 tupleDescriptor, &isNull);
-	Datum shardState = heap_getattr(heapTuple, Anum_pg_dist_shard_placement_shardstate,
+	Datum shardState = heap_getattr(heapTuple, Anum_pg_dist_placement_shardstate,
 									tupleDescriptor, &isNull);
-	Datum nodeName = heap_getattr(heapTuple, Anum_pg_dist_shard_placement_nodename,
-								  tupleDescriptor, &isNull);
-	Datum nodePort = heap_getattr(heapTuple, Anum_pg_dist_shard_placement_nodeport,
-								  tupleDescriptor, &isNull);
-	if (HeapTupleHeaderGetNatts(heapTuple->t_data) != Natts_pg_dist_shard_placement ||
+	Datum groupId = heap_getattr(heapTuple, Anum_pg_dist_placement_groupid,
+								 tupleDescriptor, &isNull);
+	if (HeapTupleHeaderGetNatts(heapTuple->t_data) != Natts_pg_dist_placement ||
 		HeapTupleHasNulls(heapTuple))
 	{
-		ereport(ERROR, (errmsg("unexpected null in pg_dist_shard_placement_tuple")));
+		ereport(ERROR, (errmsg("unexpected null in pg_dist_placement tuple")));
 	}
 
-	shardPlacement = CitusMakeNode(ShardPlacement);
+	shardPlacement = CitusMakeNode(GroupShardPlacement);
 	shardPlacement->placementId = DatumGetInt64(placementId);
 	shardPlacement->shardId = DatumGetInt64(shardId);
 	shardPlacement->shardLength = DatumGetInt64(shardLength);
 	shardPlacement->shardState = DatumGetUInt32(shardState);
-	shardPlacement->nodeName = TextDatumGetCString(nodeName);
-	shardPlacement->nodePort = DatumGetInt64(nodePort);
+	shardPlacement->groupId = DatumGetUInt32(groupId);
 
 	return shardPlacement;
 }
@@ -794,32 +844,32 @@ InsertShardRow(Oid relationId, uint64 shardId, char storageType,
 	tupleDescriptor = RelationGetDescr(pgDistShard);
 	heapTuple = heap_form_tuple(tupleDescriptor, values, isNulls);
 
-	simple_heap_insert(pgDistShard, heapTuple);
-	CatalogUpdateIndexes(pgDistShard, heapTuple);
+	CatalogTupleInsert(pgDistShard, heapTuple);
 
 	/* invalidate previous cache entry and close relation */
 	CitusInvalidateRelcacheByRelid(relationId);
 
 	CommandCounterIncrement();
-	heap_close(pgDistShard, RowExclusiveLock);
+	heap_close(pgDistShard, NoLock);
 }
 
 
 /*
  * InsertShardPlacementRow opens the shard placement system catalog, and inserts
  * a new row with the given values into that system catalog. If placementId is
- * INVALID_PLACEMENT_ID, a new placement id will be assigned.
+ * INVALID_PLACEMENT_ID, a new placement id will be assigned.Then, returns the
+ * placement id of the added shard placement.
  */
-void
+uint64
 InsertShardPlacementRow(uint64 shardId, uint64 placementId,
 						char shardState, uint64 shardLength,
-						char *nodeName, uint32 nodePort)
+						uint32 groupId)
 {
-	Relation pgDistShardPlacement = NULL;
+	Relation pgDistPlacement = NULL;
 	TupleDesc tupleDescriptor = NULL;
 	HeapTuple heapTuple = NULL;
-	Datum values[Natts_pg_dist_shard_placement];
-	bool isNulls[Natts_pg_dist_shard_placement];
+	Datum values[Natts_pg_dist_placement];
+	bool isNulls[Natts_pg_dist_placement];
 
 	/* form new shard placement tuple */
 	memset(values, 0, sizeof(values));
@@ -829,26 +879,26 @@ InsertShardPlacementRow(uint64 shardId, uint64 placementId,
 	{
 		placementId = master_get_new_placementid(NULL);
 	}
-	values[Anum_pg_dist_shard_placement_shardid - 1] = Int64GetDatum(shardId);
-	values[Anum_pg_dist_shard_placement_shardstate - 1] = CharGetDatum(shardState);
-	values[Anum_pg_dist_shard_placement_shardlength - 1] = Int64GetDatum(shardLength);
-	values[Anum_pg_dist_shard_placement_nodename - 1] = CStringGetTextDatum(nodeName);
-	values[Anum_pg_dist_shard_placement_nodeport - 1] = Int64GetDatum(nodePort);
-	values[Anum_pg_dist_shard_placement_placementid - 1] = Int64GetDatum(placementId);
+	values[Anum_pg_dist_placement_placementid - 1] = Int64GetDatum(placementId);
+	values[Anum_pg_dist_placement_shardid - 1] = Int64GetDatum(shardId);
+	values[Anum_pg_dist_placement_shardstate - 1] = CharGetDatum(shardState);
+	values[Anum_pg_dist_placement_shardlength - 1] = Int64GetDatum(shardLength);
+	values[Anum_pg_dist_placement_groupid - 1] = Int64GetDatum(groupId);
 
 	/* open shard placement relation and insert new tuple */
-	pgDistShardPlacement = heap_open(DistShardPlacementRelationId(), RowExclusiveLock);
+	pgDistPlacement = heap_open(DistPlacementRelationId(), RowExclusiveLock);
 
-	tupleDescriptor = RelationGetDescr(pgDistShardPlacement);
+	tupleDescriptor = RelationGetDescr(pgDistPlacement);
 	heapTuple = heap_form_tuple(tupleDescriptor, values, isNulls);
 
-	simple_heap_insert(pgDistShardPlacement, heapTuple);
-	CatalogUpdateIndexes(pgDistShardPlacement, heapTuple);
+	CatalogTupleInsert(pgDistPlacement, heapTuple);
 
 	CitusInvalidateRelcacheByShardId(shardId);
 
 	CommandCounterIncrement();
-	heap_close(pgDistShardPlacement, RowExclusiveLock);
+	heap_close(pgDistPlacement, NoLock);
+
+	return placementId;
 }
 
 
@@ -898,8 +948,8 @@ InsertIntoPgDistPartition(Oid relationId, char distributionMethod,
 	newTuple = heap_form_tuple(RelationGetDescr(pgDistPartition), newValues, newNulls);
 
 	/* finally insert tuple, build index entries & register cache invalidation */
-	simple_heap_insert(pgDistPartition, newTuple);
-	CatalogUpdateIndexes(pgDistPartition, newTuple);
+	CatalogTupleInsert(pgDistPartition, newTuple);
+
 	CitusInvalidateRelcacheByRelid(relationId);
 
 	RecordDistributedRelationDependencies(relationId, (Node *) distributionColumn);
@@ -938,10 +988,6 @@ RecordDistributedRelationDependencies(Oid distributedRelationId, Node *distribut
 
 	/* dependency from table entry to extension */
 	recordDependencyOn(&relationAddr, &citusExtensionAddr, DEPENDENCY_NORMAL);
-
-	/* make sure the distribution key column/expression does not just go away */
-	recordDependencyOnSingleRelExpr(&relationAddr, distributionKey, distributedRelationId,
-									DEPENDENCY_NORMAL, DEPENDENCY_NORMAL);
 }
 
 
@@ -984,7 +1030,7 @@ DeletePartitionRow(Oid distributedRelationId)
 	/* increment the counter so that next command can see the row */
 	CommandCounterIncrement();
 
-	heap_close(pgDistPartition, RowExclusiveLock);
+	heap_close(pgDistPartition, NoLock);
 }
 
 
@@ -1031,78 +1077,59 @@ DeleteShardRow(uint64 shardId)
 	CitusInvalidateRelcacheByRelid(distributedRelationId);
 
 	CommandCounterIncrement();
-	heap_close(pgDistShard, RowExclusiveLock);
+	heap_close(pgDistShard, NoLock);
 }
 
 
 /*
- * DeleteShardPlacementRow opens the shard placement system catalog, finds the
- * first (unique) row that corresponds to the given shardId and worker node, and
- * deletes this row.
+ * DeleteShardPlacementRow opens the shard placement system catalog, finds the placement
+ * with the given placementId, and deletes it.
  */
-uint64
-DeleteShardPlacementRow(uint64 shardId, char *workerName, uint32 workerPort)
+void
+DeleteShardPlacementRow(uint64 placementId)
 {
-	Relation pgDistShardPlacement = NULL;
+	Relation pgDistPlacement = NULL;
 	SysScanDesc scanDescriptor = NULL;
-	ScanKeyData scanKey[1];
-	int scanKeyCount = 1;
+	const int scanKeyCount = 1;
+	ScanKeyData scanKey[scanKeyCount];
 	bool indexOK = true;
 	HeapTuple heapTuple = NULL;
-	bool heapTupleFound = false;
 	TupleDesc tupleDescriptor = NULL;
-	int64 placementId = INVALID_PLACEMENT_ID;
 	bool isNull = false;
+	uint64 shardId = 0;
 
-	pgDistShardPlacement = heap_open(DistShardPlacementRelationId(), RowExclusiveLock);
-	tupleDescriptor = RelationGetDescr(pgDistShardPlacement);
+	pgDistPlacement = heap_open(DistPlacementRelationId(), RowExclusiveLock);
+	tupleDescriptor = RelationGetDescr(pgDistPlacement);
 
-	ScanKeyInit(&scanKey[0], Anum_pg_dist_shard_placement_shardid,
-				BTEqualStrategyNumber, F_INT8EQ, Int64GetDatum(shardId));
+	ScanKeyInit(&scanKey[0], Anum_pg_dist_placement_placementid,
+				BTEqualStrategyNumber, F_INT8EQ, Int64GetDatum(placementId));
 
-	scanDescriptor = systable_beginscan(pgDistShardPlacement,
-										DistShardPlacementShardidIndexId(), indexOK,
+	scanDescriptor = systable_beginscan(pgDistPlacement,
+										DistPlacementPlacementidIndexId(), indexOK,
 										NULL, scanKeyCount, scanKey);
 
 	heapTuple = systable_getnext(scanDescriptor);
-	while (HeapTupleIsValid(heapTuple))
-	{
-		ShardPlacement *placement = TupleToShardPlacement(tupleDescriptor, heapTuple);
-		if (strncmp(placement->nodeName, workerName, WORKER_LENGTH) == 0 &&
-			placement->nodePort == workerPort)
-		{
-			heapTupleFound = true;
-			break;
-		}
-
-		heapTuple = systable_getnext(scanDescriptor);
-	}
-
-	/* if we couldn't find the shard placement to delete, error out */
-	if (!heapTupleFound)
+	if (heapTuple == NULL)
 	{
 		ereport(ERROR, (errmsg("could not find valid entry for shard placement "
-							   UINT64_FORMAT " on node \"%s:%u\"",
-							   shardId, workerName, workerPort)));
+							   INT64_FORMAT, placementId)));
 	}
 
-	placementId = heap_getattr(heapTuple, Anum_pg_dist_shard_placement_placementid,
-							   tupleDescriptor, &isNull);
-	if (HeapTupleHeaderGetNatts(heapTuple->t_data) != Natts_pg_dist_shard_placement ||
+	shardId = heap_getattr(heapTuple, Anum_pg_dist_placement_shardid,
+						   tupleDescriptor, &isNull);
+	if (HeapTupleHeaderGetNatts(heapTuple->t_data) != Natts_pg_dist_placement ||
 		HeapTupleHasNulls(heapTuple))
 	{
-		ereport(ERROR, (errmsg("unexpected null in pg_dist_shard_placement_tuple")));
+		ereport(ERROR, (errmsg("unexpected null in pg_dist_placement tuple")));
 	}
 
-	simple_heap_delete(pgDistShardPlacement, &heapTuple->t_self);
+	simple_heap_delete(pgDistPlacement, &heapTuple->t_self);
 	systable_endscan(scanDescriptor);
 
 	CitusInvalidateRelcacheByShardId(shardId);
 
 	CommandCounterIncrement();
-	heap_close(pgDistShardPlacement, RowExclusiveLock);
-
-	return placementId;
+	heap_close(pgDistPlacement, NoLock);
 }
 
 
@@ -1113,26 +1140,26 @@ DeleteShardPlacementRow(uint64 shardId, char *workerName, uint32 workerPort)
 void
 UpdateShardPlacementState(uint64 placementId, char shardState)
 {
-	Relation pgDistShardPlacement = NULL;
+	Relation pgDistPlacement = NULL;
 	SysScanDesc scanDescriptor = NULL;
 	ScanKeyData scanKey[1];
 	int scanKeyCount = 1;
 	bool indexOK = true;
 	HeapTuple heapTuple = NULL;
 	TupleDesc tupleDescriptor = NULL;
-	Datum values[Natts_pg_dist_shard_placement];
-	bool isnull[Natts_pg_dist_shard_placement];
-	bool replace[Natts_pg_dist_shard_placement];
+	Datum values[Natts_pg_dist_placement];
+	bool isnull[Natts_pg_dist_placement];
+	bool replace[Natts_pg_dist_placement];
 	uint64 shardId = INVALID_SHARD_ID;
 	bool colIsNull = false;
 
-	pgDistShardPlacement = heap_open(DistShardPlacementRelationId(), RowExclusiveLock);
-	tupleDescriptor = RelationGetDescr(pgDistShardPlacement);
-	ScanKeyInit(&scanKey[0], Anum_pg_dist_shard_placement_placementid,
+	pgDistPlacement = heap_open(DistPlacementRelationId(), RowExclusiveLock);
+	tupleDescriptor = RelationGetDescr(pgDistPlacement);
+	ScanKeyInit(&scanKey[0], Anum_pg_dist_placement_placementid,
 				BTEqualStrategyNumber, F_INT8EQ, Int64GetDatum(placementId));
 
-	scanDescriptor = systable_beginscan(pgDistShardPlacement,
-										DistShardPlacementPlacementidIndexId(), indexOK,
+	scanDescriptor = systable_beginscan(pgDistPlacement,
+										DistPlacementPlacementidIndexId(), indexOK,
 										NULL, scanKeyCount, scanKey);
 
 	heapTuple = systable_getnext(scanDescriptor);
@@ -1145,17 +1172,16 @@ UpdateShardPlacementState(uint64 placementId, char shardState)
 
 	memset(replace, 0, sizeof(replace));
 
-	values[Anum_pg_dist_shard_placement_shardstate - 1] = CharGetDatum(shardState);
-	isnull[Anum_pg_dist_shard_placement_shardstate - 1] = false;
-	replace[Anum_pg_dist_shard_placement_shardstate - 1] = true;
+	values[Anum_pg_dist_placement_shardstate - 1] = CharGetDatum(shardState);
+	isnull[Anum_pg_dist_placement_shardstate - 1] = false;
+	replace[Anum_pg_dist_placement_shardstate - 1] = true;
 
 	heapTuple = heap_modify_tuple(heapTuple, tupleDescriptor, values, isnull, replace);
-	simple_heap_update(pgDistShardPlacement, &heapTuple->t_self, heapTuple);
 
-	CatalogUpdateIndexes(pgDistShardPlacement, heapTuple);
+	CatalogTupleUpdate(pgDistPlacement, &heapTuple->t_self, heapTuple);
 
 	shardId = DatumGetInt64(heap_getattr(heapTuple,
-										 Anum_pg_dist_shard_placement_shardid,
+										 Anum_pg_dist_placement_shardid,
 										 tupleDescriptor, &colIsNull));
 	Assert(!colIsNull);
 	CitusInvalidateRelcacheByShardId(shardId);
@@ -1163,7 +1189,7 @@ UpdateShardPlacementState(uint64 placementId, char shardState)
 	CommandCounterIncrement();
 
 	systable_endscan(scanDescriptor);
-	heap_close(pgDistShardPlacement, NoLock);
+	heap_close(pgDistPlacement, NoLock);
 }
 
 
@@ -1217,9 +1243,8 @@ UpdateColocationGroupReplicationFactor(uint32 colocationId, int replicationFacto
 	replace[Anum_pg_dist_colocation_replicationfactor - 1] = true;
 
 	newHeapTuple = heap_modify_tuple(heapTuple, tupleDescriptor, values, isnull, replace);
-	simple_heap_update(pgDistColocation, &newHeapTuple->t_self, newHeapTuple);
 
-	CatalogUpdateIndexes(pgDistColocation, newHeapTuple);
+	CatalogTupleUpdate(pgDistColocation, &newHeapTuple->t_self, newHeapTuple);
 
 	CommandCounterIncrement();
 
