@@ -9,17 +9,20 @@
 
 #include "postgres.h"
 
+#include <float.h>
 #include <limits.h>
 
 #include "catalog/pg_type.h"
 
 #include "distributed/citus_nodefuncs.h"
 #include "distributed/citus_nodes.h"
+#include "distributed/insert_select_planner.h"
 #include "distributed/metadata_cache.h"
 #include "distributed/multi_executor.h"
 #include "distributed/multi_planner.h"
 #include "distributed/multi_logical_optimizer.h"
 #include "distributed/multi_logical_planner.h"
+#include "distributed/multi_partitioning_utils.h"
 #include "distributed/multi_physical_planner.h"
 #include "distributed/multi_master_planner.h"
 #include "distributed/multi_router_planner.h"
@@ -29,6 +32,7 @@
 #include "parser/parsetree.h"
 #include "optimizer/pathnode.h"
 #include "optimizer/planner.h"
+#include "utils/lsyscache.h"
 #include "utils/memutils.h"
 
 
@@ -50,6 +54,11 @@ static CustomScanMethods RouterCustomScanMethods = {
 	RouterCreateScan
 };
 
+static CustomScanMethods CoordinatorInsertSelectCustomScanMethods = {
+	"Citus INSERT ... SELECT via coordinator",
+	CoordinatorInsertSelectCreateScan
+};
+
 static CustomScanMethods DelayedErrorCustomScanMethods = {
 	"Citus Delayed Error",
 	DelayedErrorCreateScan
@@ -61,15 +70,15 @@ static PlannedStmt * CreateDistributedPlan(PlannedStmt *localPlan, Query *origin
 										   Query *query, ParamListInfo boundParams,
 										   PlannerRestrictionContext *
 										   plannerRestrictionContext);
-static void AssignRTEIdentities(Query *queryTree);
+static void AdjustParseTree(Query *parse, bool assignRTEIdentities,
+							bool setPartitionedTablesInherited);
 static void AssignRTEIdentity(RangeTblEntry *rangeTableEntry, int rteIdentifier);
-static Node * SerializeMultiPlan(struct MultiPlan *multiPlan);
-static MultiPlan * DeserializeMultiPlan(Node *node);
 static PlannedStmt * FinalizePlan(PlannedStmt *localPlan, MultiPlan *multiPlan);
 static PlannedStmt * FinalizeNonRouterPlan(PlannedStmt *localPlan, MultiPlan *multiPlan,
 										   CustomScan *customScan);
 static PlannedStmt * FinalizeRouterPlan(PlannedStmt *localPlan, CustomScan *customScan);
 static void CheckNodeIsDumpable(Node *node);
+static Node * CheckNodeCopyAndSerialization(Node *node);
 static List * CopyPlanParamList(List *originalPlanParamList);
 static PlannerRestrictionContext * CreateAndPushPlannerRestrictionContext(void);
 static PlannerRestrictionContext * CurrentPlannerRestrictionContext(void);
@@ -85,6 +94,8 @@ multi_planner(Query *parse, int cursorOptions, ParamListInfo boundParams)
 	bool needsDistributedPlanning = NeedsDistributedPlanning(parse);
 	Query *originalQuery = NULL;
 	PlannerRestrictionContext *plannerRestrictionContext = NULL;
+	bool assignRTEIdentities = false;
+	bool setPartitionedTablesInherited = false;
 
 	/*
 	 * standard_planner scribbles on it's input, but for deparsing we need the
@@ -93,8 +104,10 @@ multi_planner(Query *parse, int cursorOptions, ParamListInfo boundParams)
 	if (needsDistributedPlanning)
 	{
 		originalQuery = copyObject(parse);
+		assignRTEIdentities = true;
+		setPartitionedTablesInherited = false;
 
-		AssignRTEIdentities(parse);
+		AdjustParseTree(parse, assignRTEIdentities, setPartitionedTablesInherited);
 	}
 
 	/* create a restriction context and put it at the end if context list */
@@ -122,6 +135,14 @@ multi_planner(Query *parse, int cursorOptions, ParamListInfo boundParams)
 	}
 	PG_END_TRY();
 
+	if (needsDistributedPlanning)
+	{
+		assignRTEIdentities = false;
+		setPartitionedTablesInherited = true;
+
+		AdjustParseTree(parse, assignRTEIdentities, setPartitionedTablesInherited);
+	}
+
 	/* remove the context from the context list */
 	PopPlannerRestrictionContext();
 
@@ -146,19 +167,18 @@ multi_planner(Query *parse, int cursorOptions, ParamListInfo boundParams)
 
 
 /*
- * AssignRTEIdentities assigns unique identities to the
- * RTE_RELATIONs in the given query.
+ * AdjustParseTree function modifies query tree by adding RTE identities to the
+ * RTE_RELATIONs and changing inh flag and relkind of partitioned tables. We
+ * perform these operations to ensure PostgreSQL's standard planner behaves as
+ * we need.
  *
- * To be able to track individual RTEs through postgres' query
- * planning, we need to be able to figure out whether an RTE is
- * actually a copy of another, rather than a different one. We
- * simply number the RTEs starting from 1.
- *
- * Note that we're only interested in RTE_RELATIONs and thus assigning
- * identifiers to those RTEs only.
+ * Please note that, we want to avoid modifying query tree as much as possible
+ * because if PostgreSQL changes the way it uses modified fields, that may break
+ * our logic.
  */
 static void
-AssignRTEIdentities(Query *queryTree)
+AdjustParseTree(Query *queryTree, bool assignRTEIdentities,
+				bool setPartitionedTablesInherited)
 {
 	List *rangeTableList = NIL;
 	ListCell *rangeTableCell = NULL;
@@ -171,12 +191,42 @@ AssignRTEIdentities(Query *queryTree)
 	{
 		RangeTblEntry *rangeTableEntry = (RangeTblEntry *) lfirst(rangeTableCell);
 
-		if (rangeTableEntry->rtekind != RTE_RELATION)
+		/*
+		 * To be able to track individual RTEs through PostgreSQL's query
+		 * planning, we need to be able to figure out whether an RTE is
+		 * actually a copy of another, rather than a different one. We
+		 * simply number the RTEs starting from 1.
+		 *
+		 * Note that we're only interested in RTE_RELATIONs and thus assigning
+		 * identifiers to those RTEs only.
+		 */
+		if (assignRTEIdentities && rangeTableEntry->rtekind == RTE_RELATION)
 		{
-			continue;
+			AssignRTEIdentity(rangeTableEntry, rteIdentifier++);
 		}
 
-		AssignRTEIdentity(rangeTableEntry, rteIdentifier++);
+		/*
+		 * We want Postgres to behave partitioned tables as regular relations
+		 * (i.e. we do not want to expand them to their partitions). To do this
+		 * we set each distributed partitioned table's inh flag to appropriate
+		 * value before and after dropping to the standart_planner.
+		 */
+		if (IsDistributedTable(rangeTableEntry->relid) &&
+			PartitionedTable(rangeTableEntry->relid))
+		{
+			rangeTableEntry->inh = setPartitionedTablesInherited;
+
+#if (PG_VERSION_NUM >= 100000)
+			if (setPartitionedTablesInherited)
+			{
+				rangeTableEntry->relkind = RELKIND_PARTITIONED_TABLE;
+			}
+			else
+			{
+				rangeTableEntry->relkind = RELKIND_RELATION;
+			}
+#endif
+		}
 	}
 }
 
@@ -224,7 +274,7 @@ IsModifyCommand(Query *query)
 	CmdType commandType = query->commandType;
 
 	if (commandType == CMD_INSERT || commandType == CMD_UPDATE ||
-		commandType == CMD_DELETE || query->hasModifyingCTE)
+		commandType == CMD_DELETE)
 	{
 		return true;
 	}
@@ -272,9 +322,19 @@ CreateDistributedPlan(PlannedStmt *localPlan, Query *originalQuery, Query *query
 
 	if (IsModifyCommand(query))
 	{
-		/* modifications are always routed through the same planner/executor */
-		distributedPlan =
-			CreateModifyPlan(originalQuery, query, plannerRestrictionContext);
+		EnsureModificationsCanRun();
+
+		if (InsertSelectIntoDistributedTable(originalQuery))
+		{
+			distributedPlan =
+				CreateInsertSelectPlan(originalQuery, plannerRestrictionContext);
+		}
+		else
+		{
+			/* modifications are always routed through the same planner/executor */
+			distributedPlan =
+				CreateModifyPlan(originalQuery, query, plannerRestrictionContext);
+		}
 
 		Assert(distributedPlan);
 	}
@@ -397,60 +457,24 @@ CreateDistributedPlan(PlannedStmt *localPlan, Query *originalQuery, Query *query
 MultiPlan *
 GetMultiPlan(CustomScan *customScan)
 {
+	Node *node = NULL;
 	MultiPlan *multiPlan = NULL;
 
 	Assert(list_length(customScan->custom_private) == 1);
 
-	multiPlan = DeserializeMultiPlan(linitial(customScan->custom_private));
+	node = (Node *) linitial(customScan->custom_private);
+	Assert(CitusIsA(node, MultiPlan));
 
-	return multiPlan;
-}
+	node = CheckNodeCopyAndSerialization(node);
 
-
-/*
- * SerializeMultiPlan returns the string representing the distributed plan in a
- * Const node.
- *
- * Note that this should be improved for 9.6+, we we can copy trees efficiently.
- * I.e. we should introduce copy support for relevant node types, and just
- * return the MultiPlan as-is for 9.6.
- */
-static Node *
-SerializeMultiPlan(MultiPlan *multiPlan)
-{
-	char *serializedMultiPlan = NULL;
-	Const *multiPlanData = NULL;
-
-	serializedMultiPlan = CitusNodeToString(multiPlan);
-
-	multiPlanData = makeNode(Const);
-	multiPlanData->consttype = CSTRINGOID;
-	multiPlanData->constlen = strlen(serializedMultiPlan);
-	multiPlanData->constvalue = CStringGetDatum(serializedMultiPlan);
-	multiPlanData->constbyval = false;
-	multiPlanData->location = -1;
-
-	return (Node *) multiPlanData;
-}
-
-
-/*
- * DeserializeMultiPlan returns the deserialized distributed plan from the string
- * representation in a Const node.
- */
-static MultiPlan *
-DeserializeMultiPlan(Node *node)
-{
-	Const *multiPlanData = NULL;
-	char *serializedMultiPlan = NULL;
-	MultiPlan *multiPlan = NULL;
-
-	Assert(IsA(node, Const));
-	multiPlanData = (Const *) node;
-	serializedMultiPlan = DatumGetCString(multiPlanData->constvalue);
-
-	multiPlan = (MultiPlan *) CitusStringToNode(serializedMultiPlan);
-	Assert(CitusIsA(multiPlan, MultiPlan));
+	/*
+	 * When using prepared statements the same plan gets reused across
+	 * multiple statements and transactions. We make several modifications
+	 * to the MultiPlan during execution such as assigning task placements
+	 * and evaluating functions and parameters. These changes should not
+	 * persist, so we always work on a copy.
+	 */
+	multiPlan = (MultiPlan *) copyObject(node);
 
 	return multiPlan;
 }
@@ -493,6 +517,12 @@ FinalizePlan(PlannedStmt *localPlan, MultiPlan *multiPlan)
 			break;
 		}
 
+		case MULTI_EXECUTOR_COORDINATOR_INSERT_SELECT:
+		{
+			customScan->methods = &CoordinatorInsertSelectCustomScanMethods;
+			break;
+		}
+
 		default:
 		{
 			customScan->methods = &DelayedErrorCustomScanMethods;
@@ -500,12 +530,13 @@ FinalizePlan(PlannedStmt *localPlan, MultiPlan *multiPlan)
 		}
 	}
 
-	multiPlanData = SerializeMultiPlan(multiPlan);
+	multiPlan->relationIdList = localPlan->relationOids;
+
+	multiPlanData = (Node *) multiPlan;
 
 	customScan->custom_private = list_make1(multiPlanData);
 	customScan->flags = CUSTOMPATH_SUPPORT_BACKWARD_SCAN;
 
-	/* check if we have a master query */
 	if (multiPlan->masterQuery)
 	{
 		finalPlan = FinalizeNonRouterPlan(localPlan, multiPlan, customScan);
@@ -633,19 +664,49 @@ RemoteScanRangeTableEntry(List *columnNameList)
 
 /*
  * CheckNodeIsDumpable checks that the passed node can be dumped using
- * CitusNodeToString(). As this checks is expensive, it's only active when
+ * nodeToString(). As this checks is expensive, it's only active when
  * assertions are enabled.
  */
 static void
 CheckNodeIsDumpable(Node *node)
 {
 #ifdef USE_ASSERT_CHECKING
-	char *out = CitusNodeToString(node);
+	char *out = nodeToString(node);
 	pfree(out);
 #endif
 }
 
 
+/*
+ * CheckNodeCopyAndSerialization checks copy/dump/read functions
+ * for nodes and returns copy of the input.
+ *
+ * It is only active when assertions are enabled, otherwise it returns
+ * the input directly. We use this to confirm that our serialization
+ * and copy logic produces the correct plan during regression tests.
+ *
+ * It does not check string equality on node dumps due to differences
+ * in some Postgres types.
+ */
+static Node *
+CheckNodeCopyAndSerialization(Node *node)
+{
+#ifdef USE_ASSERT_CHECKING
+	char *out = nodeToString(node);
+	Node *deserializedNode = (Node *) stringToNode(out);
+	Node *nodeCopy = copyObject(deserializedNode);
+	char *outCopy = nodeToString(nodeCopy);
+
+	pfree(out);
+	pfree(outCopy);
+
+	return nodeCopy;
+#else
+	return node;
+#endif
+}
+
+
 /*
  * multi_join_restriction_hook is a hook called by postgresql standard planner
  * to notify us about various planning information regarding joins. We use
@@ -688,6 +749,8 @@ multi_join_restriction_hook(PlannerInfo *root,
 	joinRestriction->joinType = jointype;
 	joinRestriction->joinRestrictInfoList = restrictInfoList;
 	joinRestriction->plannerInfo = root;
+	joinRestriction->innerrel = innerrel;
+	joinRestriction->outerrel = outerrel;
 
 	joinRestrictionContext->joinRestrictionList =
 		lappend(joinRestrictionContext->joinRestrictionList, joinRestriction);
