@@ -18,12 +18,12 @@
 
 #include "access/stratnum.h"
 #include "catalog/pg_type.h"
+#include "distributed/listutils.h"
 #include "distributed/metadata_cache.h"
 #include "distributed/master_metadata_utility.h"
 #include "distributed/multi_join_order.h"
 #include "distributed/multi_physical_planner.h"
 #include "distributed/resource_lock.h"
-#include "distributed/test_helper_functions.h" /* IWYU pragma: keep */
 #include "distributed/shard_pruning.h"
 #include "nodes/pg_list.h"
 #include "nodes/primnodes.h"
