@@ -9,6 +9,7 @@
 #include "postgres.h"
 
 #include "distributed/citus_clauses.h"
+#include "distributed/insert_select_planner.h"
 #include "distributed/multi_router_planner.h"
 
 #include "catalog/pg_type.h"
@@ -31,6 +32,7 @@ typedef struct FunctionEvaluationContext
 
 
 /* private function declarations */
+static void EvaluateValuesListsItems(List *valuesLists, PlanState *planState);
 static Node * EvaluateNodeIfReferencesFunction(Node *expression, PlanState *planState);
 static Node * PartiallyEvaluateExpressionMutator(Node *expression,
 												 FunctionEvaluationContext *context);
@@ -62,14 +64,19 @@ RequiresMasterEvaluation(Query *query)
 	{
 		RangeTblEntry *rte = (RangeTblEntry *) lfirst(rteCell);
 
-		if (rte->rtekind != RTE_SUBQUERY)
+		if (rte->rtekind == RTE_SUBQUERY)
 		{
-			continue;
+			if (RequiresMasterEvaluation(rte->subquery))
+			{
+				return true;
+			}
 		}
-
-		if (RequiresMasterEvaluation(rte->subquery))
+		else if (rte->rtekind == RTE_VALUES)
 		{
-			return true;
+			if (contain_mutable_functions((Node *) rte->values_lists))
+			{
+				return true;
+			}
 		}
 	}
 
@@ -99,12 +106,10 @@ RequiresMasterEvaluation(Query *query)
 void
 ExecuteMasterEvaluableFunctions(Query *query, PlanState *planState)
 {
-	CmdType commandType = query->commandType;
 	ListCell *targetEntryCell = NULL;
 	ListCell *rteCell = NULL;
 	ListCell *cteCell = NULL;
 	Node *modifiedNode = NULL;
-	bool insertSelectQuery = InsertSelectQuery(query);
 
 	if (query->jointree && query->jointree->quals)
 	{
@@ -122,16 +127,8 @@ ExecuteMasterEvaluableFunctions(Query *query, PlanState *planState)
 			continue;
 		}
 
-		if (commandType == CMD_INSERT && !insertSelectQuery)
-		{
-			modifiedNode = EvaluateNodeIfReferencesFunction((Node *) targetEntry->expr,
-															planState);
-		}
-		else
-		{
-			modifiedNode = PartiallyEvaluateExpression((Node *) targetEntry->expr,
-													   planState);
-		}
+		modifiedNode = PartiallyEvaluateExpression((Node *) targetEntry->expr,
+												   planState);
 
 		targetEntry->expr = (Expr *) modifiedNode;
 	}
@@ -140,12 +137,14 @@ ExecuteMasterEvaluableFunctions(Query *query, PlanState *planState)
 	{
 		RangeTblEntry *rte = (RangeTblEntry *) lfirst(rteCell);
 
-		if (rte->rtekind != RTE_SUBQUERY)
+		if (rte->rtekind == RTE_SUBQUERY)
 		{
-			continue;
+			ExecuteMasterEvaluableFunctions(rte->subquery, planState);
+		}
+		else if (rte->rtekind == RTE_VALUES)
+		{
+			EvaluateValuesListsItems(rte->values_lists, planState);
 		}
-
-		ExecuteMasterEvaluableFunctions(rte->subquery, planState);
 	}
 
 	foreach(cteCell, query->cteList)
@@ -157,6 +156,35 @@ ExecuteMasterEvaluableFunctions(Query *query, PlanState *planState)
 }
 
 
+/*
+ * EvaluateValuesListsItems siply does the work of walking over each expression
+ * in each value list contained in a multi-row INSERT's VALUES RTE. Basically
+ * a nested for loop to perform an in-place replacement of expressions with
+ * their ultimate values, should evaluation be necessary.
+ */
+static void
+EvaluateValuesListsItems(List *valuesLists, PlanState *planState)
+{
+	ListCell *exprListCell = NULL;
+
+	foreach(exprListCell, valuesLists)
+	{
+		List *exprList = (List *) lfirst(exprListCell);
+		ListCell *exprCell = NULL;
+
+		foreach(exprCell, exprList)
+		{
+			Expr *expr = (Expr *) lfirst(exprCell);
+			Node *modifiedNode = NULL;
+
+			modifiedNode = PartiallyEvaluateExpression((Node *) expr, planState);
+
+			exprCell->data.ptr_value = (void *) modifiedNode;
+		}
+	}
+}
+
+
 /*
  * Walks the expression evaluating any node which invokes a function as long as a Var
  * doesn't show up in the parameter list.
@@ -236,76 +264,36 @@ PartiallyEvaluateExpressionMutator(Node *expression, FunctionEvaluationContext *
 static Node *
 EvaluateNodeIfReferencesFunction(Node *expression, PlanState *planState)
 {
-	if (IsA(expression, FuncExpr))
+	if (expression == NULL || IsA(expression, Const))
 	{
-		FuncExpr *expr = (FuncExpr *) expression;
-
-		return (Node *) citus_evaluate_expr((Expr *) expr,
-											expr->funcresulttype,
-											exprTypmod((Node *) expr),
-											expr->funccollid,
-											planState);
+		return expression;
 	}
 
-	if (IsA(expression, OpExpr) ||
-		IsA(expression, DistinctExpr) ||
-		IsA(expression, NullIfExpr))
+	switch (nodeTag(expression))
 	{
-		/* structural equivalence */
-		OpExpr *expr = (OpExpr *) expression;
+		case T_FuncExpr:
+		case T_OpExpr:
+		case T_DistinctExpr:
+		case T_NullIfExpr:
+		case T_CoerceViaIO:
+		case T_ArrayCoerceExpr:
+		case T_ScalarArrayOpExpr:
+		case T_RowCompareExpr:
+		case T_Param:
+		case T_RelabelType:
+		case T_CoerceToDomain:
+		{
+			return (Node *) citus_evaluate_expr((Expr *) expression,
+												exprType(expression),
+												exprTypmod(expression),
+												exprCollation(expression),
+												planState);
+		}
 
-		return (Node *) citus_evaluate_expr((Expr *) expr,
-											expr->opresulttype, -1,
-											expr->opcollid,
-											planState);
-	}
-
-	if (IsA(expression, CoerceViaIO))
-	{
-		CoerceViaIO *expr = (CoerceViaIO *) expression;
-
-		return (Node *) citus_evaluate_expr((Expr *) expr,
-											expr->resulttype, -1,
-											expr->resultcollid,
-											planState);
-	}
-
-	if (IsA(expression, ArrayCoerceExpr))
-	{
-		ArrayCoerceExpr *expr = (ArrayCoerceExpr *) expression;
-
-		return (Node *) citus_evaluate_expr((Expr *) expr,
-											expr->resulttype,
-											expr->resulttypmod,
-											expr->resultcollid,
-											planState);
-	}
-
-	if (IsA(expression, ScalarArrayOpExpr))
-	{
-		ScalarArrayOpExpr *expr = (ScalarArrayOpExpr *) expression;
-
-		return (Node *) citus_evaluate_expr((Expr *) expr, BOOLOID, -1, InvalidOid,
-											planState);
-	}
-
-	if (IsA(expression, RowCompareExpr))
-	{
-		RowCompareExpr *expr = (RowCompareExpr *) expression;
-
-		return (Node *) citus_evaluate_expr((Expr *) expr, BOOLOID, -1, InvalidOid,
-											planState);
-	}
-
-	if (IsA(expression, Param))
-	{
-		Param *param = (Param *) expression;
-
-		return (Node *) citus_evaluate_expr((Expr *) param,
-											param->paramtype,
-											param->paramtypmod,
-											param->paramcollid,
-											planState);
+		default:
+		{
+			break;
+		}
 	}
 
 	return expression;
@@ -364,7 +352,11 @@ citus_evaluate_expr(Expr *expr, Oid result_type, int32 result_typmod,
 	/*
 	 * And evaluate it.
 	 */
+#if (PG_VERSION_NUM >= 100000)
+	const_val = ExecEvalExprSwitchContext(exprstate, econtext, &const_is_null);
+#else
 	const_val = ExecEvalExprSwitchContext(exprstate, econtext, &const_is_null, NULL);
+#endif
 
 	/* Get info needed about result datatype */
 	get_typlenbyval(result_type, &resultTypLen, &resultTypByVal);
