@@ -14,14 +14,11 @@
 #include "access/htup.h"
 #include "access/htup_details.h"
 #include "access/skey.h"
-#if (PG_VERSION_NUM >= 90500 && PG_VERSION_NUM < 90600)
-#include "access/stratnum.h"
-#else
 #include "access/skey.h"
-#endif
 #include "access/tupmacs.h"
 #include "access/xact.h"
 #include "catalog/indexing.h"
+#include "catalog/namespace.h"
 #include "commands/sequence.h"
 #include "distributed/colocation_utils.h"
 #include "distributed/connection_management.h"
@@ -39,6 +36,7 @@
 #include "distributed/worker_transaction.h"
 #include "lib/stringinfo.h"
 #include "storage/bufmgr.h"
+#include "storage/lmgr.h"
 #include "storage/lock.h"
 #include "storage/fd.h"
 #include "utils/builtins.h"
@@ -51,13 +49,16 @@
 /* default group size */
 int GroupSize = 1;
 
+/* config variable managed via guc.c */
+char *CurrentCluster = "default";
 
 /* local function forward declarations */
 static Datum ActivateNode(char *nodeName, int nodePort);
 static void RemoveNodeFromCluster(char *nodeName, int32 nodePort);
 static Datum AddNodeMetadata(char *nodeName, int32 nodePort, int32 groupId,
 							 char *nodeRack, bool hasMetadata, bool isActive,
-							 bool *nodeAlreadyExists);
+							 Oid nodeRole, char *nodeCluster, bool *nodeAlreadyExists);
+static uint32 CountPrimariesWithMetadata();
 static void SetNodeState(char *nodeName, int32 nodePort, bool isActive);
 static HeapTuple GetNodeTuple(char *nodeName, int32 nodePort);
 static Datum GenerateNodeTuple(WorkerNode *workerNode);
@@ -65,7 +66,8 @@ static int32 GetNextGroupId(void);
 static uint32 GetMaxGroupId(void);
 static int GetNextNodeId(void);
 static void InsertNodeRow(int nodeid, char *nodename, int32 nodeport, uint32 groupId,
-						  char *nodeRack, bool hasMetadata, bool isActive);
+						  char *nodeRack, bool hasMetadata, bool isActive, Oid nodeRole,
+						  char *nodeCluster);
 static void DeleteNodeRow(char *nodename, int32 nodeport);
 static List * ParseWorkerNodeFileAndRename(void);
 static WorkerNode * TupleToWorkerNode(TupleDesc tupleDescriptor, HeapTuple heapTuple);
@@ -73,6 +75,7 @@ static WorkerNode * TupleToWorkerNode(TupleDesc tupleDescriptor, HeapTuple heapT
 /* declarations for dynamic loading */
 PG_FUNCTION_INFO_V1(master_add_node);
 PG_FUNCTION_INFO_V1(master_add_inactive_node);
+PG_FUNCTION_INFO_V1(master_add_secondary_node);
 PG_FUNCTION_INFO_V1(master_remove_node);
 PG_FUNCTION_INFO_V1(master_disable_node);
 PG_FUNCTION_INFO_V1(master_activate_node);
@@ -90,7 +93,9 @@ master_add_node(PG_FUNCTION_ARGS)
 	text *nodeName = PG_GETARG_TEXT_P(0);
 	int32 nodePort = PG_GETARG_INT32(1);
 	char *nodeNameString = text_to_cstring(nodeName);
-	int32 groupId = 0;
+	int32 groupId = PG_GETARG_INT32(2);
+	Oid nodeRole = InvalidOid;
+	char *nodeClusterString = NULL;
 	char *nodeRack = WORKER_DEFAULT_RACK;
 	bool hasMetadata = false;
 	bool isActive = false;
@@ -99,8 +104,26 @@ master_add_node(PG_FUNCTION_ARGS)
 
 	CheckCitusVersion(ERROR);
 
+	/*
+	 * During tests this function is called before nodeRole and nodeCluster have been
+	 * created.
+	 */
+	if (PG_NARGS() == 3)
+	{
+		nodeRole = InvalidOid;
+		nodeClusterString = "default";
+	}
+	else
+	{
+		Name nodeClusterName = PG_GETARG_NAME(4);
+		nodeClusterString = NameStr(*nodeClusterName);
+
+		nodeRole = PG_GETARG_OID(3);
+	}
+
 	nodeRecord = AddNodeMetadata(nodeNameString, nodePort, groupId, nodeRack,
-								 hasMetadata, isActive, &nodeAlreadyExists);
+								 hasMetadata, isActive, nodeRole, nodeClusterString,
+								 &nodeAlreadyExists);
 
 	/*
 	 * After adding new node, if the node did not already exist, we will activate
@@ -112,7 +135,7 @@ master_add_node(PG_FUNCTION_ARGS)
 		nodeRecord = ActivateNode(nodeNameString, nodePort);
 	}
 
-	PG_RETURN_CSTRING(nodeRecord);
+	PG_RETURN_DATUM(nodeRecord);
 }
 
 
@@ -127,7 +150,10 @@ master_add_inactive_node(PG_FUNCTION_ARGS)
 	text *nodeName = PG_GETARG_TEXT_P(0);
 	int32 nodePort = PG_GETARG_INT32(1);
 	char *nodeNameString = text_to_cstring(nodeName);
-	int32 groupId = 0;
+	int32 groupId = PG_GETARG_INT32(2);
+	Oid nodeRole = PG_GETARG_OID(3);
+	Name nodeClusterName = PG_GETARG_NAME(4);
+	char *nodeClusterString = NameStr(*nodeClusterName);
 	char *nodeRack = WORKER_DEFAULT_RACK;
 	bool hasMetadata = false;
 	bool isActive = false;
@@ -137,9 +163,45 @@ master_add_inactive_node(PG_FUNCTION_ARGS)
 	CheckCitusVersion(ERROR);
 
 	nodeRecord = AddNodeMetadata(nodeNameString, nodePort, groupId, nodeRack,
-								 hasMetadata, isActive, &nodeAlreadyExists);
+								 hasMetadata, isActive, nodeRole, nodeClusterString,
+								 &nodeAlreadyExists);
 
-	PG_RETURN_CSTRING(nodeRecord);
+	PG_RETURN_DATUM(nodeRecord);
+}
+
+
+/*
+ * master_add_secondary_node adds a new secondary node to the cluster. It accepts as
+ * arguments the primary node it should share a group with.
+ */
+Datum
+master_add_secondary_node(PG_FUNCTION_ARGS)
+{
+	text *nodeName = PG_GETARG_TEXT_P(0);
+	int32 nodePort = PG_GETARG_INT32(1);
+	char *nodeNameString = text_to_cstring(nodeName);
+
+	text *primaryName = PG_GETARG_TEXT_P(2);
+	int32 primaryPort = PG_GETARG_INT32(3);
+	char *primaryNameString = text_to_cstring(primaryName);
+	int32 groupId = GroupForNode(primaryNameString, primaryPort);
+
+	Oid nodeRole = SecondaryNodeRoleId();
+	Name nodeClusterName = PG_GETARG_NAME(4);
+	char *nodeClusterString = NameStr(*nodeClusterName);
+	char *nodeRack = WORKER_DEFAULT_RACK;
+	bool hasMetadata = false;
+	bool isActive = true;
+	bool nodeAlreadyExists = false;
+	Datum nodeRecord;
+
+	CheckCitusVersion(ERROR);
+
+	nodeRecord = AddNodeMetadata(nodeNameString, nodePort, groupId, nodeRack,
+								 hasMetadata, isActive, nodeRole, nodeClusterString,
+								 &nodeAlreadyExists);
+
+	PG_RETURN_DATUM(nodeRecord);
 }
 
 
@@ -149,8 +211,8 @@ master_add_inactive_node(PG_FUNCTION_ARGS)
  * The call to the master_remove_node should be done by the super user and the specified
  * node should not have any active placements.
  * This function also deletes all reference table placements belong to the given node from
- * pg_dist_shard_placement, but it does not drop actual placement at the node. In the case
- * of re-adding the node, master_add_node first drops and re-creates the reference tables.
+ * pg_dist_placement, but it does not drop actual placement at the node. In the case of
+ * re-adding the node, master_add_node first drops and re-creates the reference tables.
  */
 Datum
 master_remove_node(PG_FUNCTION_ARGS)
@@ -168,31 +230,47 @@ master_remove_node(PG_FUNCTION_ARGS)
 
 
 /*
- * master_disable_node function sets isactive value of the provided node as inactive
- * at master node and all nodes with metadata regardless of the node having an
- * active shard placement.
+ * master_disable_node function sets isactive value of the provided node as inactive at
+ * master node and all nodes with metadata regardless of the node having an active shard
+ * placement.
+ *
  * The call to the master_disable_node must be done by the super user.
- * This function also deletes all reference table placements belong to the given
- * node from pg_dist_shard_placement, but it does not drop actual placement at
- * the node. In the case of re-activating the node, master_add_node first drops
- * and re-creates the reference tables.
+ *
+ * This function also deletes all reference table placements belong to the given node
+ * from pg_dist_placement, but it does not drop actual placement at the node. In the case
+ * of re-activating the node, master_add_node first drops and re-creates the reference
+ * tables.
  */
 Datum
 master_disable_node(PG_FUNCTION_ARGS)
 {
+	const bool onlyConsiderActivePlacements = true;
 	text *nodeNameText = PG_GETARG_TEXT_P(0);
 	int32 nodePort = PG_GETARG_INT32(1);
 
 	char *nodeName = text_to_cstring(nodeNameText);
-	bool hasShardPlacements = false;
 	bool isActive = false;
 
+	WorkerNode *workerNode = NULL;
+
 	CheckCitusVersion(ERROR);
 
-	DeleteAllReferenceTablePlacementsFromNode(nodeName, nodePort);
+	/* take an exclusive lock on pg_dist_node to serialize pg_dist_node changes */
+	LockRelationOid(DistNodeRelationId(), ExclusiveLock);
 
-	hasShardPlacements = NodeHasActiveShardPlacements(nodeName, nodePort);
-	if (hasShardPlacements)
+	workerNode = FindWorkerNodeAnyCluster(nodeName, nodePort);
+	if (workerNode == NULL)
+	{
+		ereport(ERROR, (errmsg("node at \"%s:%u\" does not exist", nodeName, nodePort)));
+	}
+
+	if (WorkerNodeIsPrimary(workerNode))
+	{
+		DeleteAllReferenceTablePlacementsFromNodeGroup(workerNode->groupId);
+	}
+
+	if (WorkerNodeIsPrimary(workerNode) &&
+		NodeGroupHasShardPlacements(workerNode->groupId, onlyConsiderActivePlacements))
 	{
 		ereport(NOTICE, (errmsg("Node %s:%d has active shard placements. Some queries "
 								"may fail after this operation. Use "
@@ -224,7 +302,124 @@ master_activate_node(PG_FUNCTION_ARGS)
 
 	nodeRecord = ActivateNode(nodeNameString, nodePort);
 
-	PG_RETURN_CSTRING(nodeRecord);
+	PG_RETURN_DATUM(nodeRecord);
+}
+
+
+/*
+ * GroupForNode returns the group which a given node belongs to.
+ *
+ * It only works if the requested node is a part of CurrentCluster.
+ */
+uint32
+GroupForNode(char *nodeName, int nodePort)
+{
+	WorkerNode *workerNode = FindWorkerNode(nodeName, nodePort);
+
+	if (workerNode == NULL)
+	{
+		ereport(ERROR, (errmsg("node at \"%s:%u\" does not exist", nodeName, nodePort)));
+	}
+
+	return workerNode->groupId;
+}
+
+
+/*
+ * WorkerNodeIsPrimary returns whether the argument represents a primary node.
+ */
+bool
+WorkerNodeIsPrimary(WorkerNode *worker)
+{
+	Oid primaryRole = PrimaryNodeRoleId();
+
+	/* if nodeRole does not yet exist, all nodes are primary nodes */
+	if (primaryRole == InvalidOid)
+	{
+		return true;
+	}
+
+	return worker->nodeRole == primaryRole;
+}
+
+
+/*
+ * WorkerNodeIsSecondary returns whether the argument represents a secondary node.
+ */
+bool
+WorkerNodeIsSecondary(WorkerNode *worker)
+{
+	Oid secondaryRole = SecondaryNodeRoleId();
+
+	/* if nodeRole does not yet exist, all nodes are primary nodes */
+	if (secondaryRole == InvalidOid)
+	{
+		return false;
+	}
+
+	return worker->nodeRole == secondaryRole;
+}
+
+
+/*
+ * WorkerNodeIsReadable returns whether we're allowed to send SELECT queries to this
+ * node.
+ */
+bool
+WorkerNodeIsReadable(WorkerNode *workerNode)
+{
+	if (ReadFromSecondaries == USE_SECONDARY_NODES_NEVER &&
+		WorkerNodeIsPrimary(workerNode))
+	{
+		return true;
+	}
+
+	if (ReadFromSecondaries == USE_SECONDARY_NODES_ALWAYS &&
+		WorkerNodeIsSecondary(workerNode))
+	{
+		return true;
+	}
+
+	return false;
+}
+
+
+/*
+ * PrimaryNodeForGroup returns the (unique) primary in the specified group.
+ *
+ * If there are any nodes in the requested group and groupContainsNodes is not NULL
+ * it will set the bool groupContainsNodes references to true.
+ */
+WorkerNode *
+PrimaryNodeForGroup(uint32 groupId, bool *groupContainsNodes)
+{
+	WorkerNode *workerNode = NULL;
+	HASH_SEQ_STATUS status;
+	HTAB *workerNodeHash = GetWorkerNodeHash();
+
+	hash_seq_init(&status, workerNodeHash);
+
+	while ((workerNode = hash_seq_search(&status)) != NULL)
+	{
+		uint32 workerNodeGroupId = workerNode->groupId;
+		if (workerNodeGroupId != groupId)
+		{
+			continue;
+		}
+
+		if (groupContainsNodes != NULL)
+		{
+			*groupContainsNodes = true;
+		}
+
+		if (WorkerNodeIsPrimary(workerNode))
+		{
+			hash_seq_term(&status);
+			return workerNode;
+		}
+	}
+
+	return NULL;
 }
 
 
@@ -236,32 +431,24 @@ master_activate_node(PG_FUNCTION_ARGS)
 static Datum
 ActivateNode(char *nodeName, int nodePort)
 {
-	Relation pgDistNode = heap_open(DistNodeRelationId(), AccessShareLock);
-	HeapTuple heapTuple = GetNodeTuple(nodeName, nodePort);
-	CommandId commandId = GetCurrentCommandId(true);
-	LockTupleMode lockTupleMode = LockTupleExclusive;
-	LockWaitPolicy lockWaitPolicy = LockWaitError;
-	bool followUpdates = false;
-	Buffer buffer = 0;
-	HeapUpdateFailureData heapUpdateFailureData;
-
 	WorkerNode *workerNode = NULL;
 	bool isActive = true;
 	Datum nodeRecord = 0;
 
-	heap_lock_tuple(pgDistNode, heapTuple, commandId, lockTupleMode, lockWaitPolicy,
-					followUpdates, &buffer, &heapUpdateFailureData);
-	ReleaseBuffer(buffer);
+	/* take an exclusive lock on pg_dist_node to serialize pg_dist_node changes */
+	LockRelationOid(DistNodeRelationId(), ExclusiveLock);
 
 	SetNodeState(nodeName, nodePort, isActive);
 
-	ReplicateAllReferenceTablesToNode(nodeName, nodePort);
+	workerNode = FindWorkerNodeAnyCluster(nodeName, nodePort);
+
+	if (WorkerNodeIsPrimary(workerNode))
+	{
+		ReplicateAllReferenceTablesToNode(nodeName, nodePort);
+	}
 
-	workerNode = FindWorkerNode(nodeName, nodePort);
 	nodeRecord = GenerateNodeTuple(workerNode);
 
-	heap_close(pgDistNode, AccessShareLock);
-
 	return nodeRecord;
 }
 
@@ -275,19 +462,31 @@ Datum
 master_initialize_node_metadata(PG_FUNCTION_ARGS)
 {
 	ListCell *workerNodeCell = NULL;
-	List *workerNodes = NULL;
+	List *workerNodes = NIL;
 	bool nodeAlreadyExists = false;
 
+	/* nodeRole and nodeCluster don't exist when this function is caled */
+	Oid nodeRole = InvalidOid;
+	char *nodeCluster = WORKER_DEFAULT_CLUSTER;
+
 	CheckCitusVersion(ERROR);
 
+	/*
+	 * This function should only ever be called from the create extension
+	 * script, but just to be sure, take an exclusive lock on pg_dist_node
+	 * to prevent concurrent calls.
+	 */
+	LockRelationOid(DistNodeRelationId(), ExclusiveLock);
+
 	workerNodes = ParseWorkerNodeFileAndRename();
+
 	foreach(workerNodeCell, workerNodes)
 	{
 		WorkerNode *workerNode = (WorkerNode *) lfirst(workerNodeCell);
 
 		AddNodeMetadata(workerNode->workerName, workerNode->workerPort, 0,
 						workerNode->workerRack, false, workerNode->isActive,
-						&nodeAlreadyExists);
+						nodeRole, nodeCluster, &nodeAlreadyExists);
 	}
 
 	PG_RETURN_BOOL(true);
@@ -361,7 +560,7 @@ get_shard_id_for_distribution_column(PG_FUNCTION_ARGS)
 		inputDataType = get_fn_expr_argtype(fcinfo->flinfo, 1);
 		distributionValueString = DatumToString(inputDatum, inputDataType);
 
-		distributionColumn = PartitionKey(relationId);
+		distributionColumn = DistPartitionKey(relationId);
 		distributionDataType = distributionColumn->vartype;
 
 		distributionValueDatum = StringToDatum(distributionValueString,
@@ -410,13 +609,40 @@ FindWorkerNode(char *nodeName, int32 nodePort)
 }
 
 
+/*
+ * FindWorkerNodeAnyCluster returns the workerNode no matter which cluster it is a part
+ * of. FindWorkerNodes, like almost every other function, acts as if nodes in other
+ * clusters do not exist.
+ */
+WorkerNode *
+FindWorkerNodeAnyCluster(char *nodeName, int32 nodePort)
+{
+	WorkerNode *workerNode = NULL;
+
+	Relation pgDistNode = heap_open(DistNodeRelationId(), AccessShareLock);
+	TupleDesc tupleDescriptor = RelationGetDescr(pgDistNode);
+
+	HeapTuple heapTuple = GetNodeTuple(nodeName, nodePort);
+	if (heapTuple != NULL)
+	{
+		workerNode = TupleToWorkerNode(tupleDescriptor, heapTuple);
+	}
+
+	heap_close(pgDistNode, NoLock);
+	return workerNode;
+}
+
+
 /*
  * ReadWorkerNodes iterates over pg_dist_node table, converts each row
  * into it's memory representation (i.e., WorkerNode) and adds them into
  * a list. Lastly, the list is returned to the caller.
+ *
+ * It skips nodes which are not in the current clusters unless requested to do otherwise
+ * by includeNodesFromOtherClusters.
  */
 List *
-ReadWorkerNodes()
+ReadWorkerNodes(bool includeNodesFromOtherClusters)
 {
 	SysScanDesc scanDescriptor = NULL;
 	ScanKeyData scanKey[1];
@@ -425,7 +651,7 @@ ReadWorkerNodes()
 	List *workerNodeList = NIL;
 	TupleDesc tupleDescriptor = NULL;
 
-	Relation pgDistNode = heap_open(DistNodeRelationId(), AccessExclusiveLock);
+	Relation pgDistNode = heap_open(DistNodeRelationId(), AccessShareLock);
 
 	scanDescriptor = systable_beginscan(pgDistNode,
 										InvalidOid, false,
@@ -437,13 +663,19 @@ ReadWorkerNodes()
 	while (HeapTupleIsValid(heapTuple))
 	{
 		WorkerNode *workerNode = TupleToWorkerNode(tupleDescriptor, heapTuple);
-		workerNodeList = lappend(workerNodeList, workerNode);
+
+		if (includeNodesFromOtherClusters ||
+			strncmp(workerNode->nodeCluster, CurrentCluster, WORKER_LENGTH) == 0)
+		{
+			/* the coordinator acts as if it never sees nodes not in it's cluster */
+			workerNodeList = lappend(workerNodeList, workerNode);
+		}
 
 		heapTuple = systable_getnext(scanDescriptor);
 	}
 
 	systable_endscan(scanDescriptor);
-	heap_close(pgDistNode, AccessExclusiveLock);
+	heap_close(pgDistNode, NoLock);
 
 	return workerNodeList;
 }
@@ -455,15 +687,15 @@ ReadWorkerNodes()
  * The call to the master_remove_node should be done by the super user. If there are
  * active shard placements on the node; the function errors out.
  * This function also deletes all reference table placements belong to the given node from
- * pg_dist_shard_placement, but it does not drop actual placement at the node. It also
+ * pg_dist_placement, but it does not drop actual placement at the node. It also
  * modifies replication factor of the colocation group of reference tables, so that
  * replication factor will be equal to worker count.
  */
 static void
 RemoveNodeFromCluster(char *nodeName, int32 nodePort)
 {
+	const bool onlyConsiderActivePlacements = false;
 	char *nodeDeleteCommand = NULL;
-	bool hasShardPlacements = false;
 	WorkerNode *workerNode = NULL;
 	List *referenceTableList = NIL;
 	uint32 deletedNodeId = INVALID_PLACEMENT_ID;
@@ -471,39 +703,53 @@ RemoveNodeFromCluster(char *nodeName, int32 nodePort)
 	EnsureCoordinator();
 	EnsureSuperUser();
 
-	workerNode = FindWorkerNode(nodeName, nodePort);
+	/* take an exclusive lock on pg_dist_node to serialize pg_dist_node changes */
+	LockRelationOid(DistNodeRelationId(), ExclusiveLock);
+
+	workerNode = FindWorkerNodeAnyCluster(nodeName, nodePort);
+	if (workerNode == NULL)
+	{
+		ereport(ERROR, (errmsg("node at \"%s:%u\" does not exist", nodeName, nodePort)));
+	}
 
 	if (workerNode != NULL)
 	{
 		deletedNodeId = workerNode->nodeId;
 	}
 
-	DeleteNodeRow(nodeName, nodePort);
+	if (WorkerNodeIsPrimary(workerNode))
+	{
+		DeleteAllReferenceTablePlacementsFromNodeGroup(workerNode->groupId);
+	}
 
-	DeleteAllReferenceTablePlacementsFromNode(nodeName, nodePort);
+	if (WorkerNodeIsPrimary(workerNode) &&
+		NodeGroupHasShardPlacements(workerNode->groupId, onlyConsiderActivePlacements))
+	{
+		ereport(ERROR, (errmsg("you cannot remove the primary node of a node group "
+							   "which has shard placements")));
+	}
+
+	DeleteNodeRow(nodeName, nodePort);
 
 	/*
 	 * After deleting reference tables placements, we will update replication factor
 	 * column for colocation group of reference tables so that replication factor will
 	 * be equal to worker count.
 	 */
-	referenceTableList = ReferenceTableOidList();
-	if (list_length(referenceTableList) != 0)
+	if (WorkerNodeIsPrimary(workerNode))
 	{
-		Oid firstReferenceTableId = linitial_oid(referenceTableList);
-		uint32 referenceTableColocationId = TableColocationId(firstReferenceTableId);
+		referenceTableList = ReferenceTableOidList();
+		if (list_length(referenceTableList) != 0)
+		{
+			Oid firstReferenceTableId = linitial_oid(referenceTableList);
+			uint32 referenceTableColocationId = TableColocationId(firstReferenceTableId);
 
-		List *workerNodeList = ActiveWorkerNodeList();
-		int workerCount = list_length(workerNodeList);
+			List *workerNodeList = ActivePrimaryNodeList();
+			int workerCount = list_length(workerNodeList);
 
-		UpdateColocationGroupReplicationFactor(referenceTableColocationId, workerCount);
-	}
-
-	hasShardPlacements = NodeHasActiveShardPlacements(nodeName, nodePort);
-	if (hasShardPlacements)
-	{
-		ereport(ERROR, (errmsg("you cannot remove a node which has active "
-							   "shard placements")));
+			UpdateColocationGroupReplicationFactor(referenceTableColocationId,
+												   workerCount);
+		}
 	}
 
 	nodeDeleteCommand = NodeDeleteCommand(deletedNodeId);
@@ -515,6 +761,30 @@ RemoveNodeFromCluster(char *nodeName, int32 nodePort)
 }
 
 
+/* CountPrimariesWithMetadata returns the number of primary nodes which have metadata. */
+static uint32
+CountPrimariesWithMetadata()
+{
+	uint32 primariesWithMetadata = 0;
+	WorkerNode *workerNode = NULL;
+
+	HASH_SEQ_STATUS status;
+	HTAB *workerNodeHash = GetWorkerNodeHash();
+
+	hash_seq_init(&status, workerNodeHash);
+
+	while ((workerNode = hash_seq_search(&status)) != NULL)
+	{
+		if (workerNode->hasMetadata && WorkerNodeIsPrimary(workerNode))
+		{
+			primariesWithMetadata++;
+		}
+	}
+
+	return primariesWithMetadata;
+}
+
+
 /*
  * AddNodeMetadata checks the given node information and adds the specified node to the
  * pg_dist_node table of the master and workers with metadata.
@@ -526,37 +796,35 @@ RemoveNodeFromCluster(char *nodeName, int32 nodePort)
  */
 static Datum
 AddNodeMetadata(char *nodeName, int32 nodePort, int32 groupId, char *nodeRack,
-				bool hasMetadata, bool isActive, bool *nodeAlreadyExists)
+				bool hasMetadata, bool isActive, Oid nodeRole, char *nodeCluster,
+				bool *nodeAlreadyExists)
 {
-	Relation pgDistNode = NULL;
 	int nextNodeIdInt = 0;
 	Datum returnData = 0;
 	WorkerNode *workerNode = NULL;
 	char *nodeDeleteCommand = NULL;
-	char *nodeInsertCommand = NULL;
-	List *workerNodeList = NIL;
+	uint32 primariesWithMetadata = 0;
 
 	EnsureCoordinator();
 	EnsureSuperUser();
 
 	*nodeAlreadyExists = false;
 
-	/* acquire a lock so that no one can do this concurrently */
-	pgDistNode = heap_open(DistNodeRelationId(), AccessExclusiveLock);
+	/*
+	 * Take an exclusive lock on pg_dist_node to serialize node changes.
+	 * We may want to relax or have more fine-grained locking in the future
+	 * to allow users to add multiple nodes concurrently.
+	 */
+	LockRelationOid(DistNodeRelationId(), ExclusiveLock);
 
-	/* check if the node already exists in the cluster */
-	workerNode = FindWorkerNode(nodeName, nodePort);
+	workerNode = FindWorkerNodeAnyCluster(nodeName, nodePort);
 	if (workerNode != NULL)
 	{
 		/* fill return data and return */
 		returnData = GenerateNodeTuple(workerNode);
-
-		/* close the heap */
-		heap_close(pgDistNode, AccessExclusiveLock);
-
 		*nodeAlreadyExists = true;
 
-		PG_RETURN_DATUM(returnData);
+		return returnData;
 	}
 
 	/* user lets Citus to decide on the group that the newly added node should be in */
@@ -574,41 +842,59 @@ AddNodeMetadata(char *nodeName, int32 nodePort, int32 groupId, char *nodeRack,
 		}
 	}
 
+	/* if nodeRole hasn't been added yet there's a constraint for one-node-per-group */
+	if (nodeRole != InvalidOid && nodeRole == PrimaryNodeRoleId())
+	{
+		WorkerNode *existingPrimaryNode = PrimaryNodeForGroup(groupId, NULL);
+
+		if (existingPrimaryNode != NULL)
+		{
+			ereport(ERROR, (errmsg("group %d already has a primary node", groupId)));
+		}
+	}
+
+	if (nodeRole == PrimaryNodeRoleId())
+	{
+		if (strncmp(nodeCluster, WORKER_DEFAULT_CLUSTER, WORKER_LENGTH) != 0)
+		{
+			ereport(ERROR, (errmsg("primaries must be added to the default cluster")));
+		}
+	}
+
 	/* generate the new node id from the sequence */
 	nextNodeIdInt = GetNextNodeId();
 
 	InsertNodeRow(nextNodeIdInt, nodeName, nodePort, groupId, nodeRack, hasMetadata,
-				  isActive);
+				  isActive, nodeRole, nodeCluster);
 
-	workerNode = FindWorkerNode(nodeName, nodePort);
+	workerNode = FindWorkerNodeAnyCluster(nodeName, nodePort);
 
-	/* send the delete command all nodes with metadata */
+	/* send the delete command to all primary nodes with metadata */
 	nodeDeleteCommand = NodeDeleteCommand(workerNode->nodeId);
 	SendCommandToWorkers(WORKERS_WITH_METADATA, nodeDeleteCommand);
 
 	/* finally prepare the insert command and send it to all primary nodes */
-	workerNodeList = list_make1(workerNode);
-	nodeInsertCommand = NodeListInsertCommand(workerNodeList);
-	SendCommandToWorkers(WORKERS_WITH_METADATA, nodeInsertCommand);
+	primariesWithMetadata = CountPrimariesWithMetadata();
+	if (primariesWithMetadata != 0)
+	{
+		List *workerNodeList = list_make1(workerNode);
+		char *nodeInsertCommand = NodeListInsertCommand(workerNodeList);
+		SendCommandToWorkers(WORKERS_WITH_METADATA, nodeInsertCommand);
+	}
 
-	heap_close(pgDistNode, AccessExclusiveLock);
-
-	/* fetch the worker node, and generate the output */
-	workerNode = FindWorkerNode(nodeName, nodePort);
 	returnData = GenerateNodeTuple(workerNode);
-
 	return returnData;
 }
 
 
 /*
  * SetNodeState function sets the isactive column of the specified worker in
- * pg_dist_node to true.
+ * pg_dist_node to isActive.
  */
 static void
 SetNodeState(char *nodeName, int32 nodePort, bool isActive)
 {
-	Relation pgDistNode = heap_open(DistNodeRelationId(), AccessShareLock);
+	Relation pgDistNode = heap_open(DistNodeRelationId(), RowExclusiveLock);
 	TupleDesc tupleDescriptor = RelationGetDescr(pgDistNode);
 	HeapTuple heapTuple = GetNodeTuple(nodeName, nodePort);
 
@@ -619,31 +905,39 @@ SetNodeState(char *nodeName, int32 nodePort, bool isActive)
 	char *nodeStateUpdateCommand = NULL;
 	WorkerNode *workerNode = NULL;
 
+	if (heapTuple == NULL)
+	{
+		ereport(ERROR, (errmsg("could not find valid entry for node \"%s:%d\"",
+							   nodeName, nodePort)));
+	}
+
 	memset(replace, 0, sizeof(replace));
 	values[Anum_pg_dist_node_isactive - 1] = BoolGetDatum(isActive);
 	isnull[Anum_pg_dist_node_isactive - 1] = false;
 	replace[Anum_pg_dist_node_isactive - 1] = true;
 
 	heapTuple = heap_modify_tuple(heapTuple, tupleDescriptor, values, isnull, replace);
-	simple_heap_update(pgDistNode, &heapTuple->t_self, heapTuple);
 
-	CatalogUpdateIndexes(pgDistNode, heapTuple);
+	CatalogTupleUpdate(pgDistNode, &heapTuple->t_self, heapTuple);
+
 	CitusInvalidateRelcacheByRelid(DistNodeRelationId());
 	CommandCounterIncrement();
 
-	heap_close(pgDistNode, AccessShareLock);
+	workerNode = TupleToWorkerNode(tupleDescriptor, heapTuple);
+
+	heap_close(pgDistNode, NoLock);
 
 	/* we also update isactive column at worker nodes */
-	workerNode = FindWorkerNode(nodeName, nodePort);
 	nodeStateUpdateCommand = NodeStateUpdateCommand(workerNode->nodeId, isActive);
 	SendCommandToWorkers(WORKERS_WITH_METADATA, nodeStateUpdateCommand);
 }
 
 
 /*
- * GetNodeTuple function returns heap tuple of given nodeName and nodePort. If
- * there are no node tuple with specified nodeName and nodePort, this function
- * errors out.
+ * GetNodeTuple function returns the heap tuple of given nodeName and nodePort. If the
+ * node is not found this function returns NULL.
+ *
+ * This function may return worker nodes from other clusters.
  */
 static HeapTuple
 GetNodeTuple(char *nodeName, int32 nodePort)
@@ -665,16 +959,13 @@ GetNodeTuple(char *nodeName, int32 nodePort)
 										NULL, scanKeyCount, scanKey);
 
 	heapTuple = systable_getnext(scanDescriptor);
-	if (!HeapTupleIsValid(heapTuple))
+	if (HeapTupleIsValid(heapTuple))
 	{
-		ereport(ERROR, (errmsg("could not find valid entry for node \"%s:%d\"",
-							   nodeName, nodePort)));
+		nodeTuple = heap_copytuple(heapTuple);
 	}
 
-	nodeTuple = heap_copytuple(heapTuple);
-
 	systable_endscan(scanDescriptor);
-	heap_close(pgDistNode, AccessShareLock);
+	heap_close(pgDistNode, NoLock);
 
 	return nodeTuple;
 }
@@ -694,6 +985,9 @@ GenerateNodeTuple(WorkerNode *workerNode)
 	Datum values[Natts_pg_dist_node];
 	bool isNulls[Natts_pg_dist_node];
 
+	Datum nodeClusterStringDatum = CStringGetDatum(workerNode->nodeCluster);
+	Datum nodeClusterNameDatum = DirectFunctionCall1(namein, nodeClusterStringDatum);
+
 	/* form new shard tuple */
 	memset(values, 0, sizeof(values));
 	memset(isNulls, false, sizeof(isNulls));
@@ -705,8 +999,9 @@ GenerateNodeTuple(WorkerNode *workerNode)
 	values[Anum_pg_dist_node_noderack - 1] = CStringGetTextDatum(workerNode->workerRack);
 	values[Anum_pg_dist_node_hasmetadata - 1] = BoolGetDatum(workerNode->hasMetadata);
 	values[Anum_pg_dist_node_isactive - 1] = BoolGetDatum(workerNode->isActive);
+	values[Anum_pg_dist_node_noderole - 1] = ObjectIdGetDatum(workerNode->nodeRole);
+	values[Anum_pg_dist_node_nodecluster - 1] = nodeClusterNameDatum;
 
-	/* open shard relation and insert new tuple */
 	pgDistNode = heap_open(DistNodeRelationId(), AccessShareLock);
 
 	/* generate the tuple */
@@ -714,8 +1009,7 @@ GenerateNodeTuple(WorkerNode *workerNode)
 	heapTuple = heap_form_tuple(tupleDescriptor, values, isNulls);
 	nodeDatum = HeapTupleGetDatum(heapTuple);
 
-	/* close the relation */
-	heap_close(pgDistNode, AccessShareLock);
+	heap_close(pgDistNode, NoLock);
 
 	return nodeDatum;
 }
@@ -837,12 +1131,16 @@ EnsureCoordinator(void)
 
 
 /*
- * InsertNodedRow opens the node system catalog, and inserts a new row with the
+ * InsertNodeRow opens the node system catalog, and inserts a new row with the
  * given values into that system catalog.
+ *
+ * NOTE: If you call this function you probably need to have taken a
+ * ShareRowExclusiveLock then checked that you're not adding a second primary to
+ * an existing group. If you don't it's possible for the metadata to become inconsistent.
  */
 static void
 InsertNodeRow(int nodeid, char *nodeName, int32 nodePort, uint32 groupId, char *nodeRack,
-			  bool hasMetadata, bool isActive)
+			  bool hasMetadata, bool isActive, Oid nodeRole, char *nodeCluster)
 {
 	Relation pgDistNode = NULL;
 	TupleDesc tupleDescriptor = NULL;
@@ -850,6 +1148,9 @@ InsertNodeRow(int nodeid, char *nodeName, int32 nodePort, uint32 groupId, char *
 	Datum values[Natts_pg_dist_node];
 	bool isNulls[Natts_pg_dist_node];
 
+	Datum nodeClusterStringDatum = CStringGetDatum(nodeCluster);
+	Datum nodeClusterNameDatum = DirectFunctionCall1(namein, nodeClusterStringDatum);
+
 	/* form new shard tuple */
 	memset(values, 0, sizeof(values));
 	memset(isNulls, false, sizeof(isNulls));
@@ -861,23 +1162,23 @@ InsertNodeRow(int nodeid, char *nodeName, int32 nodePort, uint32 groupId, char *
 	values[Anum_pg_dist_node_noderack - 1] = CStringGetTextDatum(nodeRack);
 	values[Anum_pg_dist_node_hasmetadata - 1] = BoolGetDatum(hasMetadata);
 	values[Anum_pg_dist_node_isactive - 1] = BoolGetDatum(isActive);
+	values[Anum_pg_dist_node_noderole - 1] = ObjectIdGetDatum(nodeRole);
+	values[Anum_pg_dist_node_nodecluster - 1] = nodeClusterNameDatum;
 
-	/* open shard relation and insert new tuple */
-	pgDistNode = heap_open(DistNodeRelationId(), AccessExclusiveLock);
+	pgDistNode = heap_open(DistNodeRelationId(), RowExclusiveLock);
 
 	tupleDescriptor = RelationGetDescr(pgDistNode);
 	heapTuple = heap_form_tuple(tupleDescriptor, values, isNulls);
 
-	simple_heap_insert(pgDistNode, heapTuple);
-	CatalogUpdateIndexes(pgDistNode, heapTuple);
-
-	/* close relation and invalidate previous cache entry */
-	heap_close(pgDistNode, AccessExclusiveLock);
+	CatalogTupleInsert(pgDistNode, heapTuple);
 
 	CitusInvalidateRelcacheByRelid(DistNodeRelationId());
 
 	/* increment the counter so that next command can see the row */
 	CommandCounterIncrement();
+
+	/* close relation */
+	heap_close(pgDistNode, NoLock);
 }
 
 
@@ -894,7 +1195,7 @@ DeleteNodeRow(char *nodeName, int32 nodePort)
 	SysScanDesc heapScan = NULL;
 	ScanKeyData scanKey[scanKeyCount];
 
-	Relation pgDistNode = heap_open(DistNodeRelationId(), AccessExclusiveLock);
+	Relation pgDistNode = heap_open(DistNodeRelationId(), RowExclusiveLock);
 
 	ScanKeyInit(&scanKey[0], Anum_pg_dist_node_nodename,
 				BTEqualStrategyNumber, F_TEXTEQ, CStringGetTextDatum(nodeName));
@@ -915,13 +1216,14 @@ DeleteNodeRow(char *nodeName, int32 nodePort)
 	simple_heap_delete(pgDistNode, &(heapTuple->t_self));
 
 	systable_endscan(heapScan);
-	heap_close(pgDistNode, AccessExclusiveLock);
 
 	/* ensure future commands don't use the node we just removed */
 	CitusInvalidateRelcacheByRelid(DistNodeRelationId());
 
 	/* increment the counter so that next command won't see the row */
 	CommandCounterIncrement();
+
+	heap_close(pgDistNode, NoLock);
 }
 
 
@@ -1065,14 +1367,14 @@ ParseWorkerNodeFileAndRename()
 		workerNodeList = lappend(workerNodeList, workerNode);
 	}
 
-	FreeFile(workerFileStream);
-	free(workerFilePath);
-
 	/* rename the file, marking that it is not used anymore */
 	appendStringInfo(renamedWorkerFilePath, "%s", workerFilePath);
 	appendStringInfo(renamedWorkerFilePath, ".obsolete");
 	rename(workerFilePath, renamedWorkerFilePath->data);
 
+	FreeFile(workerFileStream);
+	free(workerFilePath);
+
 	return workerNodeList;
 }
 
@@ -1102,6 +1404,10 @@ TupleToWorkerNode(TupleDesc tupleDescriptor, HeapTuple heapTuple)
 									 tupleDescriptor, &isNull);
 	Datum isActive = heap_getattr(heapTuple, Anum_pg_dist_node_isactive,
 								  tupleDescriptor, &isNull);
+	Datum nodeRole = heap_getattr(heapTuple, Anum_pg_dist_node_noderole,
+								  tupleDescriptor, &isNull);
+	Datum nodeCluster = heap_getattr(heapTuple, Anum_pg_dist_node_nodecluster,
+									 tupleDescriptor, &isNull);
 
 	Assert(!HeapTupleHasNulls(heapTuple));
 
@@ -1113,6 +1419,23 @@ TupleToWorkerNode(TupleDesc tupleDescriptor, HeapTuple heapTuple)
 	strlcpy(workerNode->workerRack, TextDatumGetCString(nodeRack), WORKER_LENGTH);
 	workerNode->hasMetadata = DatumGetBool(hasMetadata);
 	workerNode->isActive = DatumGetBool(isActive);
+	workerNode->nodeRole = DatumGetObjectId(nodeRole);
+
+	{
+		Name nodeClusterName = DatumGetName(nodeCluster);
+		char *nodeClusterString = NameStr(*nodeClusterName);
+
+		/*
+		 * nodeClusterString can be null if nodecluster column is not present.
+		 * In the case of extension creation/upgrade, master_initialize_node_metadata
+		 * function is called before the nodecluster column is added to pg_dist_node
+		 * table.
+		 */
+		if (nodeClusterString != NULL)
+		{
+			strlcpy(workerNode->nodeCluster, nodeClusterString, NAMEDATALEN);
+		}
+	}
 
 	return workerNode;
 }
