@@ -15,6 +15,8 @@
 #include "catalog/pg_collation.h"
 #include "catalog/pg_type.h"
 #include "distributed/metadata_cache.h"
+#include "distributed/multi_planner.h"
+#include "distributed/shard_pruning.h"
 #include "distributed/shardinterval_utils.h"
 #include "distributed/pg_dist_partition.h"
 #include "distributed/worker_protocol.h"
@@ -22,10 +24,9 @@
 #include "utils/memutils.h"
 
 
-static ShardInterval * SearchCachedShardInterval(Datum partitionColumnValue,
-												 ShardInterval **shardIntervalCache,
-												 int shardCount,
-												 FmgrInfo *compareFunction);
+static int SearchCachedShardInterval(Datum partitionColumnValue,
+									 ShardInterval **shardIntervalCache,
+									 int shardCount, FmgrInfo *compareFunction);
 
 
 /*
@@ -133,83 +134,166 @@ CompareShardIntervalsById(const void *leftElement, const void *rightElement)
 
 
 /*
- * FindShardIntervalIndex finds index of given shard in sorted shard interval array. For
- * this purpose, it calculates hash value of a number in its range(e.g. min value) and
- * finds which shard should contain the hashed value. Therefore this function only works
- * for hash distributed tables.
+ * CompareRelationShards is a comparison function for sorting relation
+ * to shard mappings by their relation ID and then shard ID.
  */
 int
-FindShardIntervalIndex(ShardInterval *shardInterval)
+CompareRelationShards(const void *leftElement, const void *rightElement)
 {
+	RelationShard *leftRelationShard = *((RelationShard **) leftElement);
+	RelationShard *rightRelationShard = *((RelationShard **) rightElement);
+	Oid leftRelationId = leftRelationShard->relationId;
+	Oid rightRelationId = rightRelationShard->relationId;
+	int64 leftShardId = leftRelationShard->shardId;
+	int64 rightShardId = rightRelationShard->shardId;
+
+	if (leftRelationId > rightRelationId)
+	{
+		return 1;
+	}
+	else if (leftRelationId < rightRelationId)
+	{
+		return -1;
+	}
+	else if (leftShardId > rightShardId)
+	{
+		return 1;
+	}
+	else if (leftShardId < rightShardId)
+	{
+		return -1;
+	}
+	else
+	{
+		return 0;
+	}
+}
+
+
+/*
+ * ShardIndex finds the index of given shard in sorted shard interval array.
+ *
+ * For hash partitioned tables, it calculates hash value of a number in its
+ * range (e.g. min value) and finds which shard should contain the hashed
+ * value. For reference tables, it simply returns 0. For distribution methods
+ * other than hash and reference, the function errors out.
+ */
+int
+ShardIndex(ShardInterval *shardInterval)
+{
+	int shardIndex = INVALID_SHARD_INDEX;
 	Oid distributedTableId = shardInterval->relationId;
+	Datum shardMinValue = shardInterval->minValue;
+
 	DistTableCacheEntry *cacheEntry = DistributedTableCacheEntry(distributedTableId);
 	char partitionMethod = cacheEntry->partitionMethod;
-	int shardCount = 0;
-	int32 shardMinValue = 0;
-	uint64 hashTokenIncrement = 0;
-	int shardIndex = -1;
 
 	/*
-	 * We can support it for other types of partitioned tables with simple binary scan
-	 * but it is not necessary at the moment. If we need that simply check algorithm in
-	 * FindShardInterval and SearchCachedShardInterval.
+	 * Note that, we can also support append and range distributed tables, but
+	 * currently it is not required.
 	 */
-	if (partitionMethod != DISTRIBUTE_BY_HASH)
+	if (partitionMethod != DISTRIBUTE_BY_HASH && partitionMethod != DISTRIBUTE_BY_NONE)
 	{
 		ereport(ERROR, (errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
-						errmsg("finding index of given shard is not supported for "
-							   "non-hash partitioned tables")));
+						errmsg("finding index of a given shard is only supported for "
+							   "hash distributed and reference tables")));
 	}
 
-	shardCount = cacheEntry->shardIntervalArrayLength;
-	shardMinValue = DatumGetInt32(shardInterval->minValue);
-	hashTokenIncrement = HASH_TOKEN_COUNT / shardCount;
-	shardIndex = (uint32) (shardMinValue - INT32_MIN) / hashTokenIncrement;
-
-	Assert(shardIndex <= shardCount);
-
-	/*
-	 * If the shard count is not power of 2, the range of the last
-	 * shard becomes larger than others. For that extra piece of range,
-	 * we still need to use the last shard.
-	 */
-	if (shardIndex == shardCount)
+	/* short-circuit for reference tables */
+	if (partitionMethod == DISTRIBUTE_BY_NONE)
 	{
-		shardIndex = shardCount - 1;
+		/* reference tables has only a single shard, so the index is fixed to 0 */
+		shardIndex = 0;
+
+		return shardIndex;
 	}
 
+	shardIndex = FindShardIntervalIndex(shardMinValue, cacheEntry);
+
 	return shardIndex;
 }
 
 
 /*
  * FindShardInterval finds a single shard interval in the cache for the
- * given partition column value.
+ * given partition column value. Note that reference tables do not have
+ * partition columns, thus, pass partitionColumnValue and compareFunction
+ * as NULL for them.
  */
 ShardInterval *
-FindShardInterval(Datum partitionColumnValue, ShardInterval **shardIntervalCache,
-				  int shardCount, char partitionMethod, FmgrInfo *compareFunction,
-				  FmgrInfo *hashFunction, bool useBinarySearch)
+FindShardInterval(Datum partitionColumnValue, DistTableCacheEntry *cacheEntry)
 {
-	ShardInterval *shardInterval = NULL;
+	Datum searchedValue = partitionColumnValue;
+	int shardIndex = INVALID_SHARD_INDEX;
+
+	if (cacheEntry->partitionMethod == DISTRIBUTE_BY_HASH)
+	{
+		searchedValue = FunctionCall1(cacheEntry->hashFunction, partitionColumnValue);
+	}
+
+	shardIndex = FindShardIntervalIndex(searchedValue, cacheEntry);
+
+	if (shardIndex == INVALID_SHARD_INDEX)
+	{
+		return NULL;
+	}
+
+	return cacheEntry->sortedShardIntervalArray[shardIndex];
+}
+
+
+/*
+ * FindShardIntervalIndex finds the index of the shard interval which covers
+ * the searched value. Note that the searched value must be the hashed value
+ * of the original value if the distribution method is hash.
+ *
+ * Note that, if the searched value can not be found for hash partitioned
+ * tables, we error out (unless there are no shards, in which case
+ * INVALID_SHARD_INDEX is returned). This should only happen if something is
+ * terribly wrong, either metadata tables are corrupted or we have a bug
+ * somewhere. Such as a hash function which returns a value not in the range
+ * of [INT32_MIN, INT32_MAX] can fire this.
+ */
+int
+FindShardIntervalIndex(Datum searchedValue, DistTableCacheEntry *cacheEntry)
+{
+	ShardInterval **shardIntervalCache = cacheEntry->sortedShardIntervalArray;
+	int shardCount = cacheEntry->shardIntervalArrayLength;
+	char partitionMethod = cacheEntry->partitionMethod;
+	FmgrInfo *compareFunction = cacheEntry->shardIntervalCompareFunction;
+	bool useBinarySearch = (partitionMethod != DISTRIBUTE_BY_HASH ||
+							!cacheEntry->hasUniformHashDistribution);
+	int shardIndex = INVALID_SHARD_INDEX;
+
+	if (shardCount == 0)
+	{
+		return INVALID_SHARD_INDEX;
+	}
 
 	if (partitionMethod == DISTRIBUTE_BY_HASH)
 	{
-		int hashedValue = DatumGetInt32(FunctionCall1(hashFunction,
-													  partitionColumnValue));
 		if (useBinarySearch)
 		{
 			Assert(compareFunction != NULL);
 
-			shardInterval = SearchCachedShardInterval(Int32GetDatum(hashedValue),
-													  shardIntervalCache, shardCount,
-													  compareFunction);
+			shardIndex = SearchCachedShardInterval(searchedValue, shardIntervalCache,
+												   shardCount, compareFunction);
+
+			/* we should always return a valid shard index for hash partitioned tables */
+			if (shardIndex == INVALID_SHARD_INDEX)
+			{
+				ereport(ERROR, (errcode(ERRCODE_DATA_EXCEPTION),
+								errmsg("cannot find shard interval"),
+								errdetail("Hash of the partition column value "
+										  "does not fall into any shards.")));
+			}
 		}
 		else
 		{
+			int hashedValue = DatumGetInt32(searchedValue);
 			uint64 hashTokenIncrement = HASH_TOKEN_COUNT / shardCount;
-			int shardIndex = (uint32) (hashedValue - INT32_MIN) / hashTokenIncrement;
 
+			shardIndex = (uint32) (hashedValue - INT32_MIN) / hashTokenIncrement;
 			Assert(shardIndex <= shardCount);
 
 			/*
@@ -221,28 +305,34 @@ FindShardInterval(Datum partitionColumnValue, ShardInterval **shardIntervalCache
 			{
 				shardIndex = shardCount - 1;
 			}
-
-			shardInterval = shardIntervalCache[shardIndex];
 		}
 	}
+	else if (partitionMethod == DISTRIBUTE_BY_NONE)
+	{
+		/* reference tables has a single shard, all values mapped to that shard */
+		Assert(shardCount == 1);
+
+		shardIndex = 0;
+	}
 	else
 	{
 		Assert(compareFunction != NULL);
 
-		shardInterval = SearchCachedShardInterval(partitionColumnValue,
-												  shardIntervalCache, shardCount,
-												  compareFunction);
+		shardIndex = SearchCachedShardInterval(searchedValue, shardIntervalCache,
+											   shardCount, compareFunction);
 	}
 
-	return shardInterval;
+	return shardIndex;
 }
 
 
 /*
- * SearchCachedShardInterval performs a binary search for a shard interval matching a
- * given partition column value and returns it.
+ * SearchCachedShardInterval performs a binary search for a shard interval
+ * matching a given partition column value and returns it's index in the cached
+ * array. If it can not find any shard interval with the given value, it returns
+ * INVALID_SHARD_INDEX.
  */
-static ShardInterval *
+static int
 SearchCachedShardInterval(Datum partitionColumnValue, ShardInterval **shardIntervalCache,
 						  int shardCount, FmgrInfo *compareFunction)
 {
@@ -273,13 +363,13 @@ SearchCachedShardInterval(Datum partitionColumnValue, ShardInterval **shardInter
 
 		if (DatumGetInt32(maxValueComparison) <= 0)
 		{
-			return shardIntervalCache[middleIndex];
+			return middleIndex;
 		}
 
 		lowerBoundIndex = middleIndex + 1;
 	}
 
-	return NULL;
+	return INVALID_SHARD_INDEX;
 }
 
 
