@@ -329,7 +329,6 @@ RemoveJobSchema(StringInfo schemaName)
 	if (OidIsValid(schemaId))
 	{
 		ObjectAddress schemaObject = { 0, 0, 0 };
-		bool showNotices = false;
 
 		bool permissionsOK = pg_namespace_ownercheck(schemaId, GetUserId());
 		if (!permissionsOK)
@@ -347,7 +346,16 @@ RemoveJobSchema(StringInfo schemaName)
 		 * can suppress notice messages that are typically displayed during
 		 * cascading deletes.
 		 */
-		deleteWhatDependsOn(&schemaObject, showNotices);
+#if (PG_VERSION_NUM >= 100000)
+		performDeletion(&schemaObject, DROP_CASCADE,
+						PERFORM_DELETION_INTERNAL |
+						PERFORM_DELETION_QUIETLY |
+						PERFORM_DELETION_SKIP_ORIGINAL |
+						PERFORM_DELETION_SKIP_EXTENSIONS);
+#else
+		deleteWhatDependsOn(&schemaObject, false);
+#endif
+
 		CommandCounterIncrement();
 
 		/* drop the empty schema */
@@ -386,7 +394,12 @@ CreateTaskTable(StringInfo schemaName, StringInfo relationName,
 
 	createStatement = CreateStatement(relation, columnDefinitionList);
 
+#if (PG_VERSION_NUM >= 100000)
+	relationObject = DefineRelation(createStatement, RELKIND_RELATION, InvalidOid, NULL,
+									NULL);
+#else
 	relationObject = DefineRelation(createStatement, RELKIND_RELATION, InvalidOid, NULL);
+#endif
 	relationId = relationObject.objectId;
 
 	Assert(relationId != InvalidOid);
@@ -510,11 +523,27 @@ CopyTaskFilesFromDirectory(StringInfo schemaName, StringInfo relationName,
 		copyStatement = CopyStatement(relation, fullFilename->data);
 		if (BinaryWorkerCopyFormat)
 		{
+#if (PG_VERSION_NUM >= 100000)
+			DefElem *copyOption = makeDefElem("format", (Node *) makeString("binary"),
+											  -1);
+#else
 			DefElem *copyOption = makeDefElem("format", (Node *) makeString("binary"));
+#endif
 			copyStatement->options = list_make1(copyOption);
 		}
 
+#if (PG_VERSION_NUM >= 100000)
+		{
+			ParseState *pstate = make_parsestate(NULL);
+			pstate->p_sourcetext = queryString;
+
+			DoCopy(pstate, copyStatement, -1, -1, &copiedRowCount);
+
+			free_parsestate(pstate);
+		}
+#else
 		DoCopy(copyStatement, queryString, &copiedRowCount);
+#endif
 		copiedRowTotal += copiedRowCount;
 		CommandCounterIncrement();
 	}
