@@ -16,10 +16,12 @@
 
 #include "postgres.h"
 #include "funcapi.h"
+#include "pgstat.h"
 
 #include <arpa/inet.h>
 #include <netinet/in.h>
 #include <sys/stat.h>
+#include <math.h>
 #include <unistd.h>
 
 #include "access/hash.h"
@@ -741,7 +743,12 @@ FileOutputStreamFlush(FileOutputStream file)
 	int written = 0;
 
 	errno = 0;
+#if (PG_VERSION_NUM >= 100000)
+	written = FileWrite(file.fileDescriptor, fileBuffer->data, fileBuffer->len,
+						PG_WAIT_IO);
+#else
 	written = FileWrite(file.fileDescriptor, fileBuffer->data, fileBuffer->len);
+#endif
 	if (written != fileBuffer->len)
 	{
 		ereport(ERROR, (errcode_for_file_access(),
