@@ -19,15 +19,16 @@
 
 #include "datatype/timestamp.h"
 #include "distributed/citus_nodes.h"
+#include "distributed/errormessage.h"
 #include "distributed/master_metadata_utility.h"
 #include "distributed/multi_logical_planner.h"
+#include "distributed/multi_planner.h"
 #include "lib/stringinfo.h"
 #include "nodes/parsenodes.h"
 #include "utils/array.h"
 
 
 /* Definitions local to the physical planner */
-#define JOBID_SEQUENCE_NAME "pg_dist_jobid_seq"
 #define ARRAY_OUT_FUNC_ID 751
 #define NON_PRUNABLE_JOIN -1
 #define RESERVED_HASHED_COLUMN_ID MaxAttrNumber
@@ -54,8 +55,14 @@ typedef enum CitusRTEKind
 	CITUS_RTE_SUBQUERY = RTE_SUBQUERY,  /* subquery in FROM */
 	CITUS_RTE_JOIN = RTE_JOIN,          /* join */
 	CITUS_RTE_FUNCTION = RTE_FUNCTION,  /* function in FROM */
+#if (PG_VERSION_NUM >= 100000)
+	CITUS_RTE_TABLEFUNC = RTE_TABLEFUNC, /* TableFunc(.., column list) */
+#endif
 	CITUS_RTE_VALUES = RTE_VALUES,      /* VALUES (<exprlist>), (<exprlist>), ... */
 	CITUS_RTE_CTE = RTE_CTE,            /* common table expr (WITH list element) */
+#if (PG_VERSION_NUM >= 100000)
+	CITUS_RTE_NAMEDTUPLESTORE = RTE_NAMEDTUPLESTORE, /* tuplestore, e.g. for triggers */
+#endif
 	CITUS_RTE_SHARD,
 	CITUS_RTE_REMOTE_QUERY
 } CitusRTEKind;
@@ -81,7 +88,8 @@ typedef enum
 	MAP_OUTPUT_FETCH_TASK = 5,
 	MERGE_FETCH_TASK = 6,
 	MODIFY_TASK = 7,
-	ROUTER_TASK = 8
+	ROUTER_TASK = 8,
+	DDL_TASK = 9
 } TaskType;
 
 
@@ -121,6 +129,7 @@ typedef struct Job
 	List *dependedJobList;
 	bool subqueryPushdown;
 	bool requiresMasterEvaluation; /* only applies to modify jobs */
+	bool deferredPruning;
 } Job;
 
 
@@ -146,6 +155,11 @@ typedef struct MapMergeJob
  * fetch tasks. We also forward declare the task execution struct here to avoid
  * including the executor header files.
  *
+ * We currently do not take replication model into account for tasks other
+ * than modifications. When it is set to REPLICATION_MODEL_2PC, the execution
+ * of the modification task is done with two-phase commit. Set it to
+ * REPLICATION_MODEL_INVALID if it is not relevant for the task.
+ *
  * NB: Changing this requires also changing _outTask in citus_outfuncs and _readTask
  * in citus_readfuncs to correctly (de)serialize this struct.
  */
@@ -169,9 +183,12 @@ typedef struct Task
 	uint64 shardId;                /* only applies to shard fetch tasks */
 	TaskExecution *taskExecution;  /* used by task tracker executor */
 	bool upsertQuery;              /* only applies to modify tasks */
+	char replicationModel;         /* only applies to modify tasks */
 
 	bool insertSelectQuery;
-	List *selectShardList;         /* only applies INSERT/SELECT tasks */
+	List *relationShardList;
+
+	List *rowValuesLists;          /* rows to use when building multi-row INSERT */
 } Task;
 
 
@@ -205,10 +222,25 @@ typedef struct JoinSequenceNode
 typedef struct MultiPlan
 {
 	CitusNode type;
+	CmdType operation;
+
+	bool hasReturning;
 	Job *workerJob;
 	Query *masterQuery;
-	char *masterTableName;
 	bool routerExecutable;
+	List *relationIdList;
+
+	/* INSERT ... SELECT via coordinator only */
+	Query *insertSelectSubquery;
+	List *insertTargetList;
+	Oid targetRelationId;
+
+	/*
+	 * NULL if this a valid plan, an error description otherwise. This will
+	 * e.g. be set if SQL features are present that a planner doesn't support,
+	 * or if prepared statement parameters prevented successful planning.
+	 */
+	DeferredErrorMessage *planningError;
 } MultiPlan;
 
 
@@ -227,17 +259,17 @@ typedef struct OperatorCacheEntry
 
 /* Config variable managed via guc.c */
 extern int TaskAssignmentPolicy;
+extern bool EnableUniqueJobIds;
+
 
 /* Function declarations for building physical plans and constructing queries */
-extern MultiPlan * MultiPhysicalPlanCreate(MultiTreeRoot *multiTree);
+extern MultiPlan * MultiPhysicalPlanCreate(MultiTreeRoot *multiTree,
+										   PlannerRestrictionContext *
+										   plannerRestrictionContext);
 extern StringInfo ShardFetchQueryString(uint64 shardId);
 extern Task * CreateBasicTask(uint64 jobId, uint32 taskId, TaskType taskType,
 							  char *queryString);
 
-/* Function declarations for shard pruning */
-extern List * PruneShardList(Oid relationId, Index tableId, List *whereClauseList,
-							 List *shardList);
-extern bool ContainsFalseClause(List *whereClauseList);
 extern OpExpr * MakeOpExpression(Var *variable, int16 strategyNumber);
 
 /*
@@ -255,6 +287,7 @@ extern Const * MakeInt4Constant(Datum constantValue);
 extern int CompareShardPlacements(const void *leftElement, const void *rightElement);
 extern bool ShardIntervalsOverlap(ShardInterval *firstInterval,
 								  ShardInterval *secondInterval);
+extern bool HasReferenceTable(Node *node);
 
 /* function declarations for Task and Task list operations */
 extern bool TasksEqual(const Task *a, const Task *b);
