@@ -13,6 +13,8 @@
 #include "nodes/plannodes.h"
 #include "nodes/relation.h"
 
+#include "distributed/citus_nodes.h"
+
 
 /* values used by jobs and tasks which do not require identifiers */
 #define INVALID_JOB_ID 0
@@ -23,6 +25,7 @@ typedef struct RelationRestrictionContext
 {
 	bool hasDistributedRelation;
 	bool hasLocalRelation;
+	bool allReferenceTables;
 	List *relationRestrictionList;
 } RelationRestrictionContext;
 
@@ -34,25 +37,56 @@ typedef struct RelationRestriction
 	RangeTblEntry *rte;
 	RelOptInfo *relOptInfo;
 	PlannerInfo *plannerInfo;
+	PlannerInfo *parentPlannerInfo;
+	List *parentPlannerParamList;
 	List *prunedShardIntervalList;
 } RelationRestriction;
 
+typedef struct JoinRestrictionContext
+{
+	List *joinRestrictionList;
+} JoinRestrictionContext;
+
+typedef struct JoinRestriction
+{
+	JoinType joinType;
+	List *joinRestrictInfoList;
+	PlannerInfo *plannerInfo;
+	RelOptInfo *innerrel;
+	RelOptInfo *outerrel;
+} JoinRestriction;
+
+typedef struct PlannerRestrictionContext
+{
+	RelationRestrictionContext *relationRestrictionContext;
+	JoinRestrictionContext *joinRestrictionContext;
+	MemoryContext memoryContext;
+} PlannerRestrictionContext;
+
+typedef struct RelationShard
+{
+	CitusNode type;
+	Oid relationId;
+	uint64 shardId;
+} RelationShard;
+
 
 extern PlannedStmt * multi_planner(Query *parse, int cursorOptions,
 								   ParamListInfo boundParams);
-
-extern bool HasCitusToplevelNode(PlannedStmt *planStatement);
-struct MultiPlan;
-extern struct MultiPlan * CreatePhysicalPlan(Query *originalQuery, Query *query,
-											 RelationRestrictionContext *
-											 restrictionContext);
-extern struct MultiPlan * GetMultiPlan(PlannedStmt *planStatement);
-extern PlannedStmt * MultiQueryContainerNode(PlannedStmt *result,
-											 struct MultiPlan *multiPlan);
+extern struct MultiPlan * GetMultiPlan(CustomScan *node);
 extern void multi_relation_restriction_hook(PlannerInfo *root, RelOptInfo *relOptInfo,
 											Index index, RangeTblEntry *rte);
-extern RelationRestrictionContext * CreateAndPushRestrictionContext(void);
-extern RelationRestrictionContext * CurrentRestrictionContext(void);
-extern void PopRestrictionContext(void);
+extern void multi_join_restriction_hook(PlannerInfo *root,
+										RelOptInfo *joinrel,
+										RelOptInfo *outerrel,
+										RelOptInfo *innerrel,
+										JoinType jointype,
+										JoinPathExtraData *extra);
+extern bool IsModifyCommand(Query *query);
+extern bool IsModifyMultiPlan(struct MultiPlan *multiPlan);
+extern RangeTblEntry * RemoteScanRangeTableEntry(List *columnNameList);
+
+
+extern int GetRTEIdentity(RangeTblEntry *rte);
 
 #endif /* MULTI_PLANNER_H */
