@@ -15,6 +15,7 @@
 #ifndef RELAY_UTILITY_H
 #define RELAY_UTILITY_H
 
+#include "fmgr.h"
 #include "lib/stringinfo.h"
 #include "nodes/nodes.h"
 
@@ -22,7 +23,7 @@
 /* Shard name and identifier related defines */
 #define SHARD_NAME_SEPARATOR '_'
 #define INVALID_SHARD_ID 0
-
+#define INVALID_PLACEMENT_ID 0
 
 /*
  * RelayFileState represents last known states of shards on a given node. We
@@ -41,8 +42,11 @@ typedef enum
 
 /* Function declarations to extend names in DDL commands */
 extern void RelayEventExtendNames(Node *parseTree, char *schemaName, uint64 shardId);
+extern void RelayEventExtendNamesForInterShardCommands(Node *parseTree,
+													   uint64 leftShardId,
+													   char *leftShardSchemaName,
+													   uint64 rightShardId,
+													   char *rightShardSchemaName);
 extern void AppendShardIdToName(char **name, uint64 shardId);
-extern void AppendShardIdToStringInfo(StringInfo name, uint64 shardId);
-
 
 #endif   /* RELAY_UTILITY_H */
