@@ -13,8 +13,11 @@
 #define SHARDINTERVAL_UTILS_H_
 
 #include "distributed/master_metadata_utility.h"
+#include "distributed/metadata_cache.h"
 #include "nodes/primnodes.h"
 
+#define INVALID_SHARD_INDEX -1
+
 /* OperatorCacheEntry contains information for each element in OperatorCache */
 typedef struct ShardIntervalCompareFunctionCacheEntry
 {
@@ -27,12 +30,12 @@ extern ShardInterval * LowestShardIntervalById(List *shardIntervalList);
 extern int CompareShardIntervals(const void *leftElement, const void *rightElement,
 								 FmgrInfo *typeCompareFunction);
 extern int CompareShardIntervalsById(const void *leftElement, const void *rightElement);
-extern int FindShardIntervalIndex(ShardInterval *shardInterval);
+extern int CompareRelationShards(const void *leftElement,
+								 const void *rightElement);
+extern int ShardIndex(ShardInterval *shardInterval);
 extern ShardInterval * FindShardInterval(Datum partitionColumnValue,
-										 ShardInterval **shardIntervalCache,
-										 int shardCount, char partitionMethod,
-										 FmgrInfo *compareFunction,
-										 FmgrInfo *hashFunction, bool useBinarySearch);
+										 DistTableCacheEntry *cacheEntry);
+extern int FindShardIntervalIndex(Datum searchedValue, DistTableCacheEntry *cacheEntry);
 extern bool SingleReplicatedTable(Oid relationId);
 
 #endif /* SHARDINTERVAL_UTILS_H_ */
