@@ -3,6 +3,7 @@
 --
 
 ALTER SEQUENCE pg_catalog.pg_dist_shardid_seq RESTART 13300000;
+ALTER SEQUENCE pg_catalog.pg_dist_placement_placementid_seq RESTART 13300000;
 
 -- create co-located tables
 SET citus.shard_count = 4;
@@ -116,21 +117,6 @@ WHERE
 
 \set VERBOSITY default
 
--- volatile functions should be disallowed
-INSERT INTO raw_events_second (user_id, value_1)
-SELECT
-  user_id, (random()*10)::int
-FROM
-  raw_events_first;
-
-INSERT INTO raw_events_second (user_id, value_1)
-WITH sub_cte AS (SELECT (random()*10)::int)
-SELECT
-  user_id, (SELECT * FROM sub_cte)
-FROM
-  raw_events_first;
-
-
 -- add one more row
 INSERT INTO raw_events_first (user_id, time) VALUES
                          (7, now());
@@ -235,7 +221,8 @@ FROM   (SELECT raw_events_second.user_id AS id,
         FROM   raw_events_first, 
                raw_events_second 
         WHERE  raw_events_first.user_id = raw_events_second.user_id) AS foo 
-GROUP  BY id; 
+GROUP  BY id
+ORDER  BY id;
 
 
 -- subquery one more level depth 
@@ -252,7 +239,8 @@ FROM   (SELECT SUM(raw_events_second.value_4) AS v4,
         FROM   raw_events_first, 
                raw_events_second 
         WHERE  raw_events_first.user_id = raw_events_second.user_id 
-        GROUP  BY raw_events_second.user_id) AS foo; 
+        GROUP  BY raw_events_second.user_id) AS foo
+ORDER  BY id;
 
 -- join between subqueries
 INSERT INTO agg_events
@@ -515,7 +503,9 @@ INSERT INTO agg_events
   FROM
     raw_events_first;
 
--- We do not support any set operations
+-- We support set operations via the coordinator
+BEGIN;
+
 INSERT INTO
   raw_events_first(user_id)
 SELECT
@@ -524,13 +514,15 @@ FROM
   ((SELECT user_id FROM raw_events_first) UNION
    (SELECT user_id FROM raw_events_second)) as foo;
 
+ROLLBACK;
+
 -- We do not support any set operations
 INSERT INTO
   raw_events_first(user_id)
   (SELECT user_id FROM raw_events_first) INTERSECT
   (SELECT user_id FROM raw_events_first);
 
--- We do not support any set operations
+-- If the query is router plannable then it is executed via the coordinator
 INSERT INTO
   raw_events_first(user_id)
 SELECT
@@ -1341,9 +1333,17 @@ ALTER TABLE raw_events_second DROP COLUMN value_4;
 INSERT INTO raw_events_first SELECT * FROM raw_events_second WHERE user_id = 100; 
 ROLLBACK;
 
--- Insert after copy is currently disallowed because of the way the 
--- transaction modification state is currently handled. Copy is also
--- rolled back.
+-- Altering a reference table and then performing an INSERT ... SELECT which
+-- joins with the reference table is not allowed, since the INSERT ... SELECT
+-- would read from the reference table over others connections than the ones
+-- that performed the DDL.
+BEGIN;
+ALTER TABLE reference_table ADD COLUMN z int;
+INSERT INTO raw_events_first (user_id)
+SELECT user_id FROM raw_events_second JOIN reference_table USING (user_id);
+ROLLBACK;
+
+-- Insert after copy is allowed
 BEGIN;
 COPY raw_events_second (user_id, value_1) FROM STDIN DELIMITER ',';
 100,100
@@ -1361,7 +1361,6 @@ INSERT INTO raw_events_first SELECT * FROM raw_events_second WHERE user_id = 101
 SELECT user_id FROM raw_events_first WHERE user_id = 101;
 ROLLBACK;
 
--- Copy after insert is currently disallowed.
 BEGIN;
 INSERT INTO raw_events_first SELECT * FROM raw_events_second;
 COPY raw_events_first (user_id, value_1) FROM STDIN DELIMITER ',';
@@ -1376,6 +1375,15 @@ COPY raw_events_first (user_id, value_1) FROM STDIN DELIMITER ',';
 \.
 ROLLBACK;
 
+-- Similarly, multi-row INSERTs will take part in transactions and reuse connections...
+BEGIN;
+INSERT INTO raw_events_first SELECT * FROM raw_events_second WHERE user_id = 100;
+COPY raw_events_first (user_id, value_1) FROM STDIN DELIMITER ',';
+104,104
+\.
+INSERT INTO raw_events_first (user_id, value_1) VALUES (105, 105), (106, 106);
+ROLLBACK;
+
 -- selecting from views works
 CREATE VIEW test_view AS SELECT * FROM raw_events_first;
 INSERT INTO raw_events_first (user_id, time, value_1, value_2, value_3, value_4) VALUES
@@ -1387,9 +1395,6 @@ INSERT INTO raw_events_first (user_id, time, value_1, value_2, value_3, value_4)
 INSERT INTO raw_events_second SELECT * FROM test_view WHERE user_id = 17 GROUP BY 1,2,3,4,5,6;
 SELECT count(*) FROM raw_events_second;
 
--- inserting into views does not
-INSERT INTO test_view SELECT * FROM raw_events_second;
-
 -- we need this in our next test
 truncate raw_events_first;
 
@@ -1609,6 +1614,8 @@ SELECT create_distributed_table('text_table', 'part_col');
 SELECT create_distributed_table('char_table','part_col');
 SELECT create_distributed_table('table_with_starts_with_defaults', 'c');
 
+SET client_min_messages TO DEBUG;
+
 INSERT INTO text_table (part_col) 
   SELECT 
     CASE WHEN part_col = 'onder' THEN 'marco'
@@ -1626,6 +1633,9 @@ INSERT INTO text_table (part_col) SELECT part_col::text from char_table;
 INSERT INTO text_table (part_col) SELECT (part_col = 'burak') is true FROM text_table;
 INSERT INTO text_table (part_col) SELECT val FROM text_table;
 INSERT INTO text_table (part_col) SELECT val::text FROM text_table;
+
+RESET client_min_messages;
+
 insert into table_with_starts_with_defaults (b,c) select b,c FROM table_with_starts_with_defaults;
 
 -- Test on partition column without native hash function 
@@ -1649,6 +1659,266 @@ INSERT INTO summary_table SELECT time, COUNT(*) FROM raw_table GROUP BY time;
 
 SELECT * FROM summary_table;
 
+-- Test INSERT ... SELECT via coordinator
+
+-- Select from constants
+TRUNCATE raw_events_first;
+
+INSERT INTO raw_events_first (user_id, value_1)
+SELECT * FROM (VALUES (1,2), (3,4), (5,6)) AS v(int,int);
+
+SELECT user_id, value_1 FROM raw_events_first ORDER BY user_id;
+
+-- Select from local functions
+TRUNCATE raw_events_first;
+
+CREATE SEQUENCE insert_select_test_seq;
+
+SET client_min_messages TO DEBUG;
+
+INSERT INTO raw_events_first (user_id, value_1, value_2)
+SELECT
+  s, nextval('insert_select_test_seq'), (random()*10)::int
+FROM
+  generate_series(1, 5) s;
+
+SELECT user_id, value_1 FROM raw_events_first ORDER BY user_id, value_1;
+
+-- ON CONFLICT is unsupported
+INSERT INTO raw_events_first (user_id, value_1)
+SELECT s, nextval('insert_select_test_seq') FROM generate_series(1, 5) s
+ON CONFLICT DO NOTHING;
+
+-- RETURNING is unsupported
+INSERT INTO raw_events_first (user_id, value_1)
+SELECT s, nextval('insert_select_test_seq') FROM generate_series(1, 5) s
+RETURNING *;
+
+RESET client_min_messages;
+
+-- INSERT ... SELECT and multi-shard SELECT in the same transaction is unsupported
+TRUNCATE raw_events_first;
+
+BEGIN;
+INSERT INTO raw_events_first (user_id, value_1)
+SELECT s, s FROM generate_series(1, 5) s;
+SELECT user_id, value_1 FROM raw_events_first;
+ROLLBACK;
+
+-- INSERT ... SELECT and single-shard SELECT in the same transaction is supported
+TRUNCATE raw_events_first;
+
+BEGIN;
+INSERT INTO raw_events_first (user_id, value_1)
+SELECT s, s FROM generate_series(1, 5) s;
+SELECT user_id, value_1 FROM raw_events_first WHERE user_id = 1;
+COMMIT;
+
+-- Select from local table
+TRUNCATE raw_events_first;
+
+CREATE TEMPORARY TABLE raw_events_first_local AS
+SELECT s AS u, 2*s AS v FROM generate_series(1, 5) s;
+
+INSERT INTO raw_events_first (user_id, value_1)
+SELECT u, v FROM raw_events_first_local;
+
+SELECT user_id, value_1 FROM raw_events_first ORDER BY user_id, value_1;
+
+-- Use columns in opposite order
+TRUNCATE raw_events_first;
+
+INSERT INTO raw_events_first (value_1, user_id)
+SELECT u, v FROM raw_events_first_local;
+
+SELECT user_id, value_1 FROM raw_events_first ORDER BY user_id, value_1;
+
+-- Set operations can work with opposite column order
+TRUNCATE raw_events_first;
+
+INSERT INTO raw_events_first (value_3, user_id)
+( SELECT v, u::bigint FROM raw_events_first_local )
+UNION ALL
+( SELECT v, u FROM raw_events_first_local );
+
+SELECT user_id, value_3 FROM raw_events_first ORDER BY user_id, value_3;
+
+-- Select from other distributed table with limit
+TRUNCATE raw_events_first;
+TRUNCATE raw_events_second;
+
+INSERT INTO raw_events_second (user_id, value_4)
+SELECT s, 3*s FROM generate_series (1,5) s;
+
+INSERT INTO raw_events_first (user_id, value_1)
+SELECT user_id, value_4 FROM raw_events_second LIMIT 5;
+
+SELECT user_id, value_1 FROM raw_events_first ORDER BY user_id, value_1;
+
+-- CTEs are supported in local queries
+TRUNCATE raw_events_first;
+
+WITH removed_rows AS (
+  DELETE FROM raw_events_first_local RETURNING u
+)
+INSERT INTO raw_events_first (user_id, value_1)
+WITH value AS (SELECT 1)
+SELECT * FROM removed_rows, value;
+
+SELECT user_id, value_1 FROM raw_events_first ORDER BY user_id, value_1;
+
+-- nested CTEs are also supported
+TRUNCATE raw_events_first;
+INSERT INTO raw_events_first_local SELECT s, 2*s FROM generate_series(0, 10) s;
+
+WITH rows_to_remove AS (
+  SELECT u FROM raw_events_first_local WHERE u > 0
+),
+removed_rows AS (
+  DELETE FROM raw_events_first_local
+  WHERE u IN (SELECT * FROM rows_to_remove)
+  RETURNING u, v
+)
+INSERT INTO raw_events_first (user_id, value_1)
+WITH ultra_rows AS (
+  WITH numbers AS (
+    SELECT s FROM generate_series(1,10) s
+  ),
+  super_rows AS (
+    SELECT u, v FROM removed_rows JOIN numbers ON (u = s)
+  )
+  SELECT * FROM super_rows LIMIT 5
+)
+SELECT u, v FROM ultra_rows;
+
+SELECT user_id, value_1 FROM raw_events_first ORDER BY user_id, value_1;
+
+-- CTEs with duplicate names are also supported
+TRUNCATE raw_events_first;
+
+WITH super_rows AS (
+  SELECT u FROM raw_events_first_local
+)
+INSERT INTO raw_events_first (user_id, value_1)
+WITH super_rows AS (
+  SELECT * FROM super_rows GROUP BY u
+)
+SELECT u, 5 FROM super_rows;
+
+SELECT user_id, value_1 FROM raw_events_first ORDER BY user_id, value_1;
+
+-- CTEs are supported in router queries
+TRUNCATE raw_events_first;
+
+WITH user_two AS (
+  SELECT user_id, value_4 FROM raw_events_second WHERE user_id = 2
+)
+INSERT INTO raw_events_first (user_id, value_1)
+SELECT * FROM user_two;
+
+SELECT user_id, value_1 FROM raw_events_first ORDER BY user_id, value_1;
+
+-- CTEs are supported when there are name collisions
+WITH numbers AS (
+  SELECT s FROM generate_series(1,10) s
+)
+INSERT INTO raw_events_first(user_id, value_1)
+WITH numbers AS (
+  SELECT s, s FROM generate_series(1,5) s
+)
+SELECT * FROM numbers;
+
+-- Select into distributed table with a sequence
+CREATE TABLE "CaseSensitiveTable" ("UserID" int, "Value1" int);
+SELECT create_distributed_table('"CaseSensitiveTable"', 'UserID');
+
+INSERT INTO "CaseSensitiveTable"
+SELECT s, s FROM generate_series(1,10) s;
+
+SELECT * FROM "CaseSensitiveTable" ORDER BY "UserID";
+
+DROP TABLE "CaseSensitiveTable";
+
+-- Select into distributed table with a sequence
+CREATE TABLE dist_table_with_sequence (user_id serial, value_1 serial);
+SELECT create_distributed_table('dist_table_with_sequence', 'user_id');
+
+-- from local query
+INSERT INTO dist_table_with_sequence (value_1)
+SELECT s FROM generate_series(1,5) s;
+
+SELECT * FROM dist_table_with_sequence ORDER BY user_id;
+
+-- from a distributed query
+INSERT INTO dist_table_with_sequence (value_1)
+SELECT value_1 FROM dist_table_with_sequence;
+
+SELECT * FROM dist_table_with_sequence ORDER BY user_id;
+
+-- Select from distributed table into reference table
+CREATE TABLE ref_table (user_id int, value_1 int);
+SELECT create_reference_table('ref_table');
+
+INSERT INTO ref_table
+SELECT user_id, value_1 FROM raw_events_second;
+
+SELECT * FROM ref_table ORDER BY user_id, value_1;
+
+DROP TABLE ref_table;
+
+-- Select into an append-partitioned table is not supported
+CREATE TABLE insert_append_table (user_id int, value_4 bigint);
+SELECT create_distributed_table('insert_append_table', 'user_id', 'append');
+
+INSERT INTO insert_append_table (user_id, value_4)
+SELECT user_id, 1 FROM raw_events_second LIMIT 5;
+
+DROP TABLE insert_append_table;
+
+-- Insert from other distributed table as prepared statement
+TRUNCATE raw_events_first;
+
+PREPARE insert_prep(int) AS
+INSERT INTO raw_events_first (user_id, value_1)
+SELECT $1, value_4 FROM raw_events_second ORDER BY value_4 LIMIT 1;
+
+EXECUTE insert_prep(1);
+EXECUTE insert_prep(2);
+EXECUTE insert_prep(3);
+EXECUTE insert_prep(4);
+EXECUTE insert_prep(5);
+EXECUTE insert_prep(6);
+
+SELECT user_id, value_1 FROM raw_events_first ORDER BY user_id, value_1;
+
+-- Inserting into views is handled via coordinator
+TRUNCATE raw_events_first;
+
+INSERT INTO test_view
+SELECT * FROM raw_events_second;
+
+SELECT user_id, value_4 FROM test_view ORDER BY user_id, value_4;
+
+-- Drop the view now, because the column we are about to drop depends on it
+DROP VIEW test_view;
+
+-- Make sure we handle dropped columns correctly
+CREATE TABLE drop_col_table (col1 text, col2 text, col3 text);
+SELECT create_distributed_table('drop_col_table', 'col2');
+
+ALTER TABLE drop_col_table DROP COLUMN col1;
+
+INSERT INTO drop_col_table (col3, col2)
+SELECT value_4, user_id FROM raw_events_second LIMIT 5;
+
+SELECT * FROM drop_col_table ORDER BY col2, col3;
+
+-- make sure the tuple went to the right shard
+SELECT * FROM drop_col_table WHERE col2 = '1';
+
+RESET client_min_messages;
+
+DROP TABLE drop_col_table;
 DROP TABLE raw_table;
 DROP TABLE summary_table;
 DROP TABLE raw_events_first CASCADE;
