@@ -7,7 +7,6 @@
 
 
 ALTER SEQUENCE pg_catalog.pg_dist_shardid_seq RESTART 700000;
-ALTER SEQUENCE pg_catalog.pg_dist_jobid_seq RESTART 700000;
 
 
 SET citus.large_table_shard_count TO 2;
@@ -15,7 +14,13 @@ SET client_min_messages TO DEBUG2;
 SET citus.task_executor_type TO 'task-tracker';
 
 -- Single range-repartition join to test join-pruning behaviour.
-
+EXPLAIN (COSTS OFF)
+SELECT
+	count(*)
+FROM
+	orders, customer
+WHERE
+	o_custkey = c_custkey;
 SELECT
 	count(*)
 FROM
@@ -25,7 +30,14 @@ WHERE
 
 -- Single range-repartition join with a selection clause on the partitioned
 -- table to test the case when all map tasks are pruned away.
-
+EXPLAIN (COSTS OFF)
+SELECT
+	count(*)
+FROM
+	orders, customer
+WHERE
+	o_custkey = c_custkey AND
+	o_orderkey < 0;
 SELECT
 	count(*)
 FROM
@@ -36,7 +48,14 @@ WHERE
 
 -- Single range-repartition join with a selection clause on the base table to
 -- test the case when all sql tasks are pruned away.
-
+EXPLAIN (COSTS OFF)
+SELECT
+	count(*)
+FROM
+	orders, customer
+WHERE
+	o_custkey = c_custkey AND
+	c_custkey < 0;
 SELECT
 	count(*)
 FROM
@@ -48,7 +67,13 @@ WHERE
 -- Dual hash-repartition join test case. Note that this query doesn't produce
 -- meaningful results and is only to test hash-partitioning of two large tables
 -- on non-partition columns.
-
+EXPLAIN (COSTS OFF)
+SELECT
+	count(*)
+FROM
+	lineitem, customer
+WHERE
+	l_partkey = c_nationkey;
 SELECT
 	count(*)
 FROM
@@ -58,7 +83,14 @@ WHERE
 
 -- Dual hash-repartition join with a selection clause on one of the tables to
 -- test the case when all map tasks are pruned away.
-
+EXPLAIN (COSTS OFF)
+SELECT
+	count(*)
+FROM
+	lineitem, customer
+WHERE
+	l_partkey = c_nationkey AND
+	l_orderkey < 0;
 SELECT
 	count(*)
 FROM
@@ -68,6 +100,14 @@ WHERE
 	l_orderkey < 0;
 
 -- Test cases with false in the WHERE clause
+EXPLAIN (COSTS OFF)
+SELECT
+	o_orderkey
+FROM
+	orders INNER JOIN customer ON (o_custkey = c_custkey)
+WHERE
+	false;
+-- execute once, to verify that's handled
 SELECT
 	o_orderkey
 FROM
@@ -75,6 +115,7 @@ FROM
 WHERE
 	false;
 
+EXPLAIN (COSTS OFF)
 SELECT
 	o_orderkey
 FROM
@@ -82,11 +123,13 @@ FROM
 WHERE
 	1=0 AND c_custkey < 0;
 
+EXPLAIN (COSTS OFF)
 SELECT
 	o_orderkey
 FROM
 	orders INNER JOIN customer ON (o_custkey = c_custkey AND false);
 
+EXPLAIN (COSTS OFF)
 SELECT
 	o_orderkey
 FROM
