@@ -10,7 +10,7 @@ SELECT create_reference_table('reference_table_test');
 
 -- see that partkey is NULL
 SELECT
-	partmethod, (partkey IS NULL) as partkeyisnull, colocationid, repmodel
+	partmethod, (partkey IS NULL) as partkeyisnull, repmodel
 FROM
 	pg_dist_partition
 WHERE
@@ -28,7 +28,9 @@ SELECT
 FROM
 	pg_dist_shard_placement
 WHERE
-	shardid IN (SELECT shardid FROM pg_dist_shard WHERE logicalrelid = 'reference_table_test'::regclass);
+	shardid IN (SELECT shardid FROM pg_dist_shard WHERE logicalrelid = 'reference_table_test'::regclass)
+ORDER BY
+	placementid;
 
 -- check whether data was copied into distributed table
 SELECT * FROM reference_table_test;
@@ -228,17 +230,26 @@ WHERE
 	value_1 = 1 OR value_1 = 2;
 
 -- set operations are supported
-(SELECT * FROM reference_table_test WHERE value_1 = 1)
-UNION
-(SELECT * FROM reference_table_test WHERE value_1 = 3);
+SELECT * FROM (
+	SELECT * FROM reference_table_test WHERE value_1 = 1
+	UNION
+	SELECT * FROM reference_table_test WHERE value_1 = 3
+) AS combination
+ORDER BY value_1;
 
-(SELECT * FROM reference_table_test WHERE value_1 = 1)
-EXCEPT
-(SELECT * FROM reference_table_test WHERE value_1 = 3);
+SELECT * FROM (
+	SELECT * FROM reference_table_test WHERE value_1 = 1
+	EXCEPT
+	SELECT * FROM reference_table_test WHERE value_1 = 3
+) AS combination
+ORDER BY value_1;
 
-(SELECT * FROM reference_table_test WHERE value_1 = 1)
-INTERSECT
-(SELECT * FROM reference_table_test WHERE value_1 = 3);
+SELECT * FROM (
+	SELECT * FROM reference_table_test WHERE value_1 = 1
+	INTERSECT
+	SELECT * FROM reference_table_test WHERE value_1 = 3
+) AS combination
+ORDER BY value_1;
 
 -- to make the tests more interested for aggregation tests, ingest some more data
 INSERT INTO reference_table_test VALUES (1, 1.0, '1', '2016-12-01');
@@ -492,7 +503,9 @@ SELECT
 FROM
 	pg_dist_shard_placement
 WHERE
-	shardid IN (SELECT shardid FROM pg_dist_shard WHERE logicalrelid = 'reference_table_test_fourth'::regclass);
+	shardid IN (SELECT shardid FROM pg_dist_shard WHERE logicalrelid = 'reference_table_test_fourth'::regclass)
+ORDER BY
+	placementid;
 
 -- let's not run some update/delete queries on arbitrary columns
 DELETE FROM
@@ -666,7 +679,6 @@ WHERE
 	colocated_table_test.value_1 = reference_table_test.value_1 AND colocated_table_test_2.value_1 = reference_table_test.value_1;
 
 
-SET client_min_messages TO NOTICE;
 SET citus.log_multi_join_order TO FALSE;
 
 SET citus.shard_count TO DEFAULT;
@@ -674,7 +686,7 @@ SET citus.task_executor_type to "real-time";
 
 -- some INSERT .. SELECT queries that involve both hash distributed and reference tables
 
--- should error out since we're inserting into reference table where 
+-- should go via coordinator since we're inserting into reference table where 
 -- not all the participants are reference tables
 INSERT INTO
 	reference_table_test (value_1)
@@ -683,9 +695,9 @@ SELECT
 FROM
 	colocated_table_test, colocated_table_test_2
 WHERE
-	colocated_table_test.value_1 = colocated_table_test.value_1;
+	colocated_table_test.value_1 = colocated_table_test_2.value_1;
 
--- should error out, same as the above
+-- should go via coordinator, same as the above
 INSERT INTO
 	reference_table_test (value_1)
 SELECT
@@ -695,8 +707,7 @@ FROM
 WHERE
 	colocated_table_test.value_1 = reference_table_test.value_1;
 
--- now, insert into the hash partitioned table and use reference 
--- tables in the SELECT queries
+-- not pushable due to lack of equality between partition column and column of reference table
 INSERT INTO
 	colocated_table_test (value_1, value_2)
 SELECT 
@@ -718,8 +729,7 @@ WHERE
 	colocated_table_test_2.value_2 = reference_table_test.value_2
 RETURNING value_1, value_2;
 
--- partition column value comes from reference table but still first error is
--- on data type mismatch
+-- partition column value comes from reference table, goes via coordinator
 INSERT INTO
 	colocated_table_test (value_1, value_2)
 SELECT
@@ -727,10 +737,8 @@ SELECT
 FROM
 	colocated_table_test_2, reference_table_test
 WHERE
-	colocated_table_test_2.value_4 = reference_table_test.value_4
-RETURNING value_1, value_2;
+	colocated_table_test_2.value_4 = reference_table_test.value_4;
 
--- partition column value comes from reference table which should error out
 INSERT INTO
 	colocated_table_test (value_1, value_2)
 SELECT
@@ -738,9 +746,9 @@ SELECT
 FROM
 	colocated_table_test_2, reference_table_test
 WHERE
-	colocated_table_test_2.value_4 = reference_table_test.value_4
-RETURNING value_1, value_2;
+	colocated_table_test_2.value_4 = reference_table_test.value_4;
 
+RESET client_min_messages;
 
 -- some tests for mark_tables_colocated
 -- should error out
@@ -850,15 +858,18 @@ ALTER TABLE reference_table_ddl ALTER COLUMN value_2 SET DEFAULT 25.0;
 ALTER TABLE reference_table_ddl ALTER COLUMN value_3 SET NOT NULL;
 
 -- see that Citus applied all DDLs to the table
-\d reference_table_ddl
+SELECT "Column", "Type", "Modifiers" FROM table_desc WHERE relid='public.reference_table_ddl'::regclass;
+\d reference_index_2
 
 -- also to the shard placements
 \c - - - :worker_1_port
-\d reference_table_ddl*
+SELECT "Column", "Type", "Modifiers" FROM table_desc WHERE relid='public.reference_table_ddl_1250019'::regclass;
+\d reference_index_2_1250019
 \c - - - :master_port
 DROP INDEX reference_index_2;
 \c - - - :worker_1_port
-\d reference_table_ddl*
+SELECT "Column", "Type", "Modifiers" FROM table_desc WHERE relid='public.reference_table_ddl_1250019'::regclass;
+\di reference_index_2*
 \c - - - :master_port
 
 -- as we expect, renaming and setting WITH OIDS does not work for reference tables
@@ -982,13 +993,13 @@ UPDATE reference_table_test SET value_1 = 10 WHERE value_1 = 2;
 COMMIT;
 SELECT * FROM reference_table_test;
 
--- do not allow mixing transactions
+-- DML+master_modify_multiple_shards is allowed
 BEGIN;
 INSERT INTO reference_table_test VALUES (2, 2.0, '2', '2016-12-02');
 SELECT master_modify_multiple_shards('DELETE FROM colocated_table_test');
 ROLLBACK;
 
--- Do not allow DDL and modification in the same transaction
+-- DDL+DML is allowed
 BEGIN;
 ALTER TABLE reference_table_test ADD COLUMN value_dummy INT;
 INSERT INTO reference_table_test VALUES (2, 2.0, '2', '2016-12-02');
