@@ -264,14 +264,16 @@ SELECT
 	id, substring(title, 2, 1) AS subtitle, count(*)
 	FROM articles_hash
 	WHERE author_id = 1 or author_id = 3
-	GROUP BY GROUPING SETS ((id),(subtitle));
+	GROUP BY GROUPING SETS ((id),(subtitle))
+	ORDER BY id, subtitle;
 
 -- grouping sets are not supported on multiple shards
 SELECT
 	id, substring(title, 2, 1) AS subtitle, count(*)
 	FROM articles_hash
 	WHERE author_id = 1 or author_id = 2
-	GROUP BY GROUPING SETS ((id),(subtitle));
+	GROUP BY GROUPING SETS ((id),(subtitle))
+	ORDER BY id, subtitle;
 
 -- queries which involve functions in FROM clause are supported if it goes to a single worker.
 SELECT * FROM articles_hash, position('om' in 'Thomas') WHERE author_id = 1;
@@ -396,16 +398,14 @@ SELECT max(word_count)
 	WHERE author_id = 1
 	GROUP BY author_id;
 
-	
--- router plannable union queries are supported
-(SELECT * FROM articles_hash WHERE author_id = 1)
-UNION
-(SELECT * FROM articles_hash WHERE author_id = 3);
 
+-- router plannable union queries are supported
 SELECT * FROM (
-	(SELECT * FROM articles_hash WHERE author_id = 1)
+	SELECT * FROM articles_hash WHERE author_id = 1
 	UNION
-	(SELECT * FROM articles_hash WHERE author_id = 3)) uu;
+	SELECT * FROM articles_hash WHERE author_id = 3
+) AS combination
+ORDER BY id;
 
 (SELECT LEFT(title, 1) FROM articles_hash WHERE author_id = 1)
 UNION
@@ -415,9 +415,12 @@ UNION
 INTERSECT
 (SELECT LEFT(title, 1) FROM articles_hash WHERE author_id = 3);
 
-(SELECT LEFT(title, 2) FROM articles_hash WHERE author_id = 1)
-EXCEPT
-(SELECT LEFT(title, 2) FROM articles_hash WHERE author_id = 3);
+SELECT * FROM (
+	SELECT LEFT(title, 2) FROM articles_hash WHERE author_id = 1
+	EXCEPT
+	SELECT LEFT(title, 2) FROM articles_hash WHERE author_id = 3
+) AS combination
+ORDER BY 1;
 
 -- union queries are not supported if not router plannable
 -- there is an inconsistency on shard pruning between
@@ -627,16 +630,24 @@ SELECT a.author_id as first_author, b.word_count as second_word_count
 	FROM articles_hash a, articles_single_shard_hash b
 	WHERE a.author_id = 10 and a.author_id = b.author_id and
 		date_ne_timestamp('1954-04-11', '1954-04-11'::timestamp);
+
 -- union/difference /intersection with where false
 -- this query was not originally router plannable, addition of 1=0
 -- makes it router plannable
-(SELECT * FROM articles_hash WHERE author_id = 1)
-UNION
-(SELECT * FROM articles_hash WHERE author_id = 2 and 1=0);
+SELECT * FROM (
+	SELECT * FROM articles_hash WHERE author_id = 1
+	UNION
+	SELECT * FROM articles_hash WHERE author_id = 2 and 1=0
+) AS combination
+ORDER BY id;
+
+SELECT * FROM (
+	SELECT * FROM articles_hash WHERE author_id = 1
+	EXCEPT
+	SELECT * FROM articles_hash WHERE author_id = 2 and 1=0
+) AS combination
+ORDER BY id;
 
-(SELECT * FROM articles_hash WHERE author_id = 1)
-EXCEPT
-(SELECT * FROM articles_hash WHERE author_id = 2 and 1=0);
 
 (SELECT * FROM articles_hash WHERE author_id = 1)
 INTERSECT
