@@ -56,8 +56,8 @@ DROP EXTENSION citus;
 CREATE EXTENSION citus;
 
 -- re-add the nodes to the cluster
-SELECT master_add_node('localhost', :worker_1_port);
-SELECT master_add_node('localhost', :worker_2_port);
+SELECT 1 FROM master_add_node('localhost', :worker_1_port);
+SELECT 1 FROM master_add_node('localhost', :worker_2_port);
 
 -- create a table with a SERIAL column
 CREATE TABLE testserialtable(id serial, group_id integer);
@@ -80,10 +80,6 @@ ALTER SEQUENCE standalone_sequence OWNED BY testserialtable.group_id;
 -- an edge case, but it's OK to change an owner to the same distributed table
 ALTER SEQUENCE testserialtable_id_seq OWNED BY testserialtable.id;
 
--- verify sequence was created on worker
-\c - - - :worker_1_port
-\ds
-
 -- drop distributed table
 \c - - - :master_port
 DROP TABLE testserialtable;
